### Definitions ###

cdef enum Window:
    BLACKMAN = 0
    HAMMING = 1
    HANNING = 2
    BARTLETT = 3
    TRAPEZOID = 4
    RECTANGULAR = 5

cdef enum Boolean:
    FA = 0
    TR = 1

cdef extern from "SPTK.h":

    # Library routines
    double _agexp "agexp"(double r, double x, double y)
    double _gexp "gexp"(const double r, const double x)
    double _glog "glog"(const double r, const double x)
    int _mseq "mseq"()


    void _acorr "acorr"(double *x, int l, double *r, const int np)

    # Adaptive mel-generalized cepstrum analysis
    double _acep "acep"(double x, double *c, const int m, const double lambda_coef,
                        const double step, const double tau, const int pd,
                        const double eps);
    double _agcep "agcep"(double x, double *c, const int m, const int stage,
                          const double lambda_coef, const double step, const double tau,
                          const double eps);
    double _amcep "amcep"(double x, double *b, const int m, const double a,
                          const double lambda_coef, const double step, const double tau,
                          const int pd, const double eps);

    # Mel-generalized cepstrum analysis
    int _mcep "mcep"(double *xw, const int flng, double *mc, const int m, const double a,
                     const int itr1, const int itr2, const double dd, const int etype,
                     const double e, const double f, const int itype)
    int _gcep "gcep"(double *xw, const int flng, double *gc, const int m, const double g,
                     const int itr1, const int itr2, const double d, const int etype,
                     const double e, const double f, const int itype)
    int _mgcep "mgcep"(double *xw, int flng, double *b, const int m, const double a,
                       const double g, const int n, const int itr1, const int itr2,
                       const double dd, const int etype, const double e, const double f,
                       const int itype)
    int _uels "uels"(double *xw, const int flng, double *c, const int m, const int itr1,
                     const int itr2, const double dd, const int etype, const double e,
                     const int itype)
    void _fftcep "fftcep"(double *sp, const int flng, double *c, const int m, int itr,
                          double ac)
    int _lpc "lpc"(double *x, const int flng, double *a, const int m, const double f)


    # MFCC
    void _mfcc "mfcc"(double *in_mfcc, double *mc, const double sampleFreq,
                      const double alpha, const double eps, const int wlng,
                      const int flng, const int m, const int n, const int ceplift,
                      const Boolean dftmode, const Boolean usehamming)

    # LPC, LSP and PARCOR conversions
    void _lpc2c "lpc2c"(double *a, int m1, double *c, const int m2)
    int _lpc2lsp "lpc2lsp"(double *lpc, double *lsp, const int order, const int numsp,
                           const int maxitr, const double eps)
    int _lpc2par "lpc2par"(double *a, double *k, const int m)
    void _par2lpc "par2lpc"(double *k, double *a, const int m)
    void _lsp2lpc "lsp2lpc"(double *lsp, double *a, const int m)
    void _lsp2sp "lsp2sp"(double *lsp, const int m, double *x, const int l, const int gain)


    # Mel-generalized cepstrum conversions
    void _mc2b "mc2b"(double *mc, double *b, int m, const double a)
    void _b2mc "b2mc"(double *b, double *mc, int m, const double a)
    void _b2c "b2c"(double *b, int m1, double *c, int m2, double a)
    void _c2acr "c2acr"(double *c, const int m1, double *r, const int m2, const int flng)
    void _c2ir "c2ir"(double *c, const int nc, double *h, const int leng)
    void _ic2ir "ic2ir"(double *h, const int leng, double *c, const int nc)
    void _c2ndps "c2ndps"(double *c, const int m, double *n, const int l)
    void _ndps2c "ndps2c"(double *n, const int l, double *c, const int m)
    void _gc2gc "gc2gc"(double *c1, const int m1, const double g1, double *c2, const int m2,
                        const double g2)
    void _gnorm "gnorm"(double *c1, double *c2, int m, const double g)
    void _ignorm "ignorm"(double *c1, double *c2, int m, const double g)
    void _freqt "freqt"(double *c1, const int m1, double *c2, const int m2, const double a)
    void _frqtr "frqtr"(double *c1, int m1, double *c2, int m2, const double a)
    void _mgc2mgc "mgc2mgc"(double *c1, const int m1, const double a1, const double g1,
                            double *c2, const int m2, const double a2, const double g2)
    void _mgc2sp "mgc2sp"(double *mgc, const int m, const double a, const double g, double *x,
                          double *y, const int flng)
    void _mgclsp2sp "mgclsp2sp"(double a, double g, double *lsp, const int m, double *x,
                                const int l, const int gain)


    # F0 analysis
    void _swipe "swipe"(double *input, double *output, int length, int samplerate,
                        int frame_shift, double min, double max, double st, int otype)
    int _rapt "rapt"(float * input, float * output, int length, double sample_freq,
                     int frame_shift, double minF0, double maxF0, double voice_bias,
                     int otype)


    # Excitation
    void _excite "excite"(double *pitch, int n, double *out, int fprd, int iprd, Boolean gauss, int seed_i)

    # Waveform generation filters
    double _zerodf "zerodf"(double x, double *b, int m, double *d)
    double _zerodft "zerodft"(double x, double *b, const int m, double *d)
    double _poledf "poledf"(double x, double *a, int m, double *d)
    double _poledft "poledft"(double x, double *a, int m, double *d)
    double _lmadf "lmadf"(double x, double *c, const int m, const int pd, double *d)
    double _lmadft "lmadft"(double x, double *c, const int m, const int pd, double *d,
                            int block_num, int *block_size)
    double _lspdf_even "lspdf_even"(double x, double *f, const int m, double *d)
    double _lspdf_odd "lspdf_odd"(double x, double *f, const int m, double *d)
    double _ltcdf "ltcdf"(double x, double *k, int m, double *d)
    double _glsadf "glsadf"(double x, double *c, const int m, const int n, double *d)
    double _glsadft "glsadft"(double x, double *c, const int m, const int n, double *d)
    double _iglsadf "iglsadf"(double x, double *c, const int m, const int n, double *d)
    double _iglsadft "iglsadft"(double x, double *c, const int m, const int n, double *d)
    double _mlsadf "mlsadf"(double x, double *b, const int m, const double a, const int pd,
                            double *d)
    double _mlsadft "mlsadft"(double x, double *b, const int m, const double a, const int pd,
                             double *d)
    double _mglsadf "mglsadf"(double x, double *b, const int m, const double a, const int n,
                              double *d)
    double _mglsadft "mglsadft"(double x, double *b, const int m, const double a, const int n,
                                double *d)
    double _imglsadf "imglsadf"(double x, double *b, const int m, const double a, const int n,
                                double *d)
    double _imglsadft "imglsadft"(double x, double *b, const int m, const double a, const int n,
                                  double *d)

    # Window functions
    double _window "window"(Window window_type, double *x, const int size, const int nflg)

    # Utils
    int _lspcheck "lspcheck"(double *lsp, const int ord)
    void _phidf "phidf"(const double x, const int m, double a, double *d)
    int _levdur "levdur"(double *r, double *a, const int m, double eps)



import numpy as np
import pysptk
import pytest
from pysptk.synthesis import Synthesizer


def __dummy_source():
    np.random.seed(98765)
    return np.random.randn(2 ** 14)


def __dummy_windowed_frames(source, frame_len=512, hopsize=80):
    np.random.seed(98765)
    n_frames = int(len(source) / hopsize) + 1
    windowed = np.random.randn(n_frames, frame_len) * pysptk.blackman(frame_len)
    return 0.5 * 32768.0 * windowed


@pytest.mark.parametrize("order", [20, 25])
@pytest.mark.parametrize("pd", [4, 5, 6, 7])
def test_LMADF(order, pd):
    from pysptk.synthesis import LMADF

    def __test_synthesis(filt):
        # dummy source excitation
        source = __dummy_source()

        hopsize = 80

        # dummy filter coef.
        windowed = __dummy_windowed_frames(source, frame_len=512, hopsize=hopsize)
        b = pysptk.mcep(windowed, filt.order, 0.0)

        # synthesis
        synthesizer = Synthesizer(filt, hopsize)
        y = synthesizer.synthesis(source, b)
        assert np.all(np.isfinite(y))

    def __test(order, pd):
        __test_synthesis(LMADF(order, pd=pd))

    __test(order, pd)


def test_LMADF_corner_case():
    from pysptk.synthesis import LMADF

    def __test_invalid_pade(pd):
        LMADF(20, pd=pd)

    with pytest.raises(ValueError):
        __test_invalid_pade(3)
    with pytest.raises(ValueError):
        __test_invalid_pade(8)


@pytest.mark.parametrize("order", [20, 25])
@pytest.mark.parametrize("pd", [4, 5, 6, 7])
@pytest.mark.parametrize("alpha", [0.0, 0.41])
def test_MLSADF(order, pd, alpha):
    def __test_synthesis(filt):
        # dummy source excitation
        source = __dummy_source()

        hopsize = 80

        # dummy filter coef.
        windowed = __dummy_windowed_frames(source, frame_len=512, hopsize=hopsize)
        mc = pysptk.mcep(windowed, filt.order, filt.alpha)
        b = pysptk.mc2b(mc, filt.alpha)

        # synthesis
        synthesizer = Synthesizer(filt, hopsize)
        y = synthesizer.synthesis(source, b)
        assert np.all(np.isfinite(y))

        # transpose
        synthesizer = Synthesizer(filt, hopsize, transpose=True)
        y = synthesizer.synthesis(source, b)
        assert np.all(np.isfinite(y))

    from pysptk.synthesis import MLSADF

    def __test(order, alpha, pd):
        __test_synthesis(MLSADF(order, alpha, pd=pd))

    __test(order, alpha, pd)


def test_MLSADF_corner_case():
    from pysptk.synthesis import MLSADF

    def __test_invalid_pade(pd):
        MLSADF(20, pd=pd)

    with pytest.raises(ValueError):
        __test_invalid_pade(3)
    with pytest.raises(ValueError):
        __test_invalid_pade(8)


@pytest.mark.parametrize("order", [20, 25])
@pytest.mark.parametrize("stage", [2, 5, 10])
def test_GLSADF(order, stage):
    from pysptk.synthesis import GLSADF

    def __test_synthesis(filt):
        # dummy source excitation
        source = __dummy_source()

        hopsize = 80

        # dummy filter coef.
        windowed = __dummy_windowed_frames(source, frame_len=512, hopsize=hopsize)
        gamma = -1.0 / filt.stage
        mgc = pysptk.mgcep(windowed, filt.order, 0.0, gamma)
        b = pysptk.mgc2b(mgc, 0.0, gamma)

        # synthesis
        synthesizer = Synthesizer(filt, hopsize)
        y = synthesizer.synthesis(source, b)
        assert np.all(np.isfinite(y))

        # transpose
        synthesizer = Synthesizer(filt, hopsize, transpose=True)
        y = synthesizer.synthesis(source, b)
        assert np.all(np.isfinite(y))

    def __test(order, stage):
        __test_synthesis(GLSADF(order, stage))

    __test(order, stage)

    def __test_invalid_stage(stage):
        GLSADF(20, stage=stage)


def test_GLSADF_corner_case():
    from pysptk.synthesis import GLSADF

    def __test_invalid_stage(stage):
        GLSADF(20, stage=stage)

    with pytest.raises(ValueError):
        __test_invalid_stage(-1)
    with pytest.raises(ValueError):
        __test_invalid_stage(0)


@pytest.mark.parametrize("order", [20, 25])
@pytest.mark.parametrize("alpha", [0.0, 0.41])
@pytest.mark.parametrize("stage", [2, 5, 10])
def test_MGLSADF(order, alpha, stage):
    from pysptk.synthesis import MGLSADF

    def __test_synthesis(filt):
        # dummy source excitation
        source = __dummy_source()

        hopsize = 80

        # dummy filter coef.
        windowed = __dummy_windowed_frames(source, frame_len=512, hopsize=hopsize)
        gamma = -1.0 / filt.stage
        mgc = pysptk.mgcep(windowed, filt.order, filt.alpha, gamma)
        b = pysptk.mgc2b(mgc, filt.alpha, gamma)

        # synthesis
        synthesizer = Synthesizer(filt, hopsize)
        y = synthesizer.synthesis(source, b)
        assert np.all(np.isfinite(y))

        # transpose
        synthesizer = Synthesizer(filt, hopsize, transpose=True)
        y = synthesizer.synthesis(source, b)
        assert np.all(np.isfinite(y))

    def __test(order, alpha, stage):
        __test_synthesis(MGLSADF(order, alpha, stage))

    __test(order, alpha, stage)


def test_MGLSADF_corner_case():
    from pysptk.synthesis import MGLSADF

    def __test_invalid_stage(stage):
        MGLSADF(20, stage=stage)

    with pytest.raises(ValueError):
        __test_invalid_stage(-1)
    with pytest.raises(ValueError):
        __test_invalid_stage(0)


def test_AllZeroDF():
    pass

    def __test_synthesis(filt):
        # dummy source excitation
        source = __dummy_source()

        hopsize = 80

        # dummy filter coef.
        windowed = __dummy_windowed_frames(source, frame_len=512, hopsize=hopsize)
        lpc = pysptk.lpc(windowed, filt.order)
        lpc[:, 0] = 0
        b = -lpc

        # synthesis
        synthesizer = Synthesizer(filt, hopsize)
        y = synthesizer.synthesis(source, b)
        assert np.all(np.isfinite(y))

        # transpose
        synthesizer = Synthesizer(filt, hopsize, transpose=True)
        y = synthesizer.synthesis(source, b)
        assert np.all(np.isfinite(y))


@pytest.mark.parametrize("order", [20, 25])
def test_AllPoleDF(order):
    from pysptk.synthesis import AllPoleDF

    def __test_synthesis(filt):
        # dummy source excitation
        source = __dummy_source()

        hopsize = 80

        # dummy filter coef.
        windowed = __dummy_windowed_frames(source, frame_len=512, hopsize=hopsize)
        lpc = pysptk.lpc(windowed, filt.order)

        # make sure lpc has loggain
        lpc[:, 0] = np.log(lpc[:, 0])

        # synthesis
        synthesizer = Synthesizer(filt, hopsize)
        y = synthesizer.synthesis(source, lpc)
        assert np.all(np.isfinite(y))

        # transpose
        synthesizer = Synthesizer(filt, hopsize, transpose=True)
        y = synthesizer.synthesis(source, lpc)
        assert np.all(np.isfinite(y))

    def __test_synthesis_levdur(filt):
        # dummy source excitation
        source = __dummy_source()

        hopsize = 80

        # dummy filter coef.
        windowed = __dummy_windowed_frames(source, frame_len=512, hopsize=hopsize)
        c = pysptk.mcep(windowed, filt.order)
        a = pysptk.c2acr(c)
        lpc = pysptk.levdur(a)
        lpc2 = pysptk.levdur(a, use_scipy=True)
        assert np.allclose(lpc, lpc2)

        # make sure lpc has loggain
        lpc[:, 0] = np.log(lpc[:, 0])

        # synthesis
        synthesizer = Synthesizer(filt, hopsize)
        y = synthesizer.synthesis(source, lpc)
        assert np.all(np.isfinite(y))

    def __test(order):
        __test_synthesis(AllPoleDF(order))
        __test_synthesis_levdur(AllPoleDF(order))

    __test(order)


@pytest.mark.parametrize("order", [20, 25])
def test_AllPoleLatticeDF(order):
    from pysptk.synthesis import AllPoleLatticeDF

    def __test_synthesis(filt):
        # dummy source excitation
        source = __dummy_source()

        hopsize = 80

        # dummy filter coef.
        windowed = __dummy_windowed_frames(source, frame_len=512, hopsize=hopsize)
        lpc = pysptk.lpc(windowed, filt.order)
        par = pysptk.lpc2par(lpc)

        # make sure par has loggain
        par[:, 0] = np.log(par[:, 0])

        # synthesis
        synthesizer = Synthesizer(filt, hopsize)
        y = synthesizer.synthesis(source, par)
        assert np.all(np.isfinite(y))

    __test_synthesis(AllPoleLatticeDF(order))


@pytest.mark.parametrize("order", [20, 25])
def test_LSPDF(order):
    from pysptk.synthesis import LSPDF

    def __test_synthesis(filt):
        # dummy source excitation
        source = __dummy_source()

        hopsize = 80

        # dummy filter coef.
        windowed = __dummy_windowed_frames(source, frame_len=512, hopsize=hopsize)
        lpc = pysptk.lpc(windowed, filt.order)
        lsp = pysptk.lpc2lsp(lpc)
        # make sure lsp has loggain
        lsp[:, 0] = np.log(lsp[:, 0])

        # synthesis
        synthesizer = Synthesizer(filt, hopsize)
        y = synthesizer.synthesis(source, lsp)
        assert np.all(np.isfinite(y))

    __test_synthesis(LSPDF(order))



# coding: utf-8

"""
A python wrapper for `Speech Signal Processing Toolkit (SPTK)
<http://sp-tk.sourceforge.net>`_.

https://github.com/r9y9/pysptk

The wrapper is based on a modified version of SPTK (`r9y9/SPTK`_)

.. _r9y9/SPTK: https://github.com/r9y9/SPTK

Full documentation
------------------

A full documentation of SPTK is available at http://sp-tk.sourceforge.net.
If you are not familiar with SPTK, I recommend you to take a look at the doc
first before using ``pysptk``.

Demonstration notebooks
-----------------------

* `Introduction notebook`_: a brief introduction to pysptk
* `Speech analysis and re-synthesis resynthesis notebook`_: a demonstration \
notebook for speech analysis and re-synthesis. Synthesized audio examples\
(English) are available on the notebook.

.. _Introduction notebook: \
http://nbviewer.ipython.org/github/r9y9/pysptk/blob/master/examples/\
pysptk%20introduction.ipynb
.. _Speech analysis and re-synthesis resynthesis notebook:\
 http://nbviewer.ipython.org/github/r9y9/pysptk/blob/master/examples/\
Speech%20analysis%20and%20re-synthesis.ipynb

"""

from . import synthesis, util  # noqa
from .conversion import mc2e, mc2sp, mgc2b, sp2mc  # noqa
from .sptk import *  # noqa
from .version import __version__  # noqa



import numpy as np
import pysptk
import pytest
from pysptk.util import apply_along_last_axis, automatic_type_conversion, mcepalpha


def test_assert_gamma():
    def __test(gamma):
        pysptk.util.assert_gamma(gamma)

    for gamma in [-2.0, 0.1]:
        with pytest.raises(ValueError):
            __test(gamma)


def test_assert_pade():
    def __test(pade):
        pysptk.util.assert_pade(pade)

    for pade in [3, 8]:
        with pytest.raises(ValueError):
            __test(pade)


def test_assert_fftlen():
    def __test(fftlen):
        pysptk.util.assert_fftlen(fftlen)

    for fftlen in [255, 257]:
        with pytest.raises(ValueError):
            __test(fftlen)


@pytest.mark.parametrize("order", [15, 20, 25, 30])
@pytest.mark.parametrize("alpha", [0.35, 0.41, 0.5])
def test_phidf(order, alpha):
    np.random.seed(98765)
    dummy_input = np.random.rand(64)
    delay = np.zeros(order + 1)
    for x in dummy_input:
        pysptk.phidf(x, order, alpha, delay)
        assert np.all(np.isfinite(delay))


@pytest.mark.parametrize("order", [15, 20, 25, 30])
def test_lspcheck(order):
    np.random.seed(98765)
    lsp = np.random.rand(order + 1)
    pysptk.lspcheck(lsp)
    # TODO: valid check


def test_example_audio_file():
    from os.path import exists

    path = pysptk.util.example_audio_file()
    assert exists(path)


def test_mcepalpha():
    assert np.isclose(mcepalpha(8000), 0.312)
    assert np.isclose(mcepalpha(11025), 0.357)
    assert np.isclose(mcepalpha(16000), 0.41)
    assert np.isclose(mcepalpha(22050), 0.455)
    assert np.isclose(mcepalpha(44100), 0.544)
    assert np.isclose(mcepalpha(48000), 0.554)


def test_automatic_type_conversion():
    @automatic_type_conversion
    def f(x):
        return x

    for dtype in [np.float32, np.float16, np.float64]:
        x = np.ones(10, dtype=dtype)
        y = f(x)
        assert y.dtype == x.dtype
        y = f(x=x)
        assert y.dtype == x.dtype


def test_apply_along_last_axis():
    @apply_along_last_axis
    def f(x):
        assert x.ndim == 1
        return x[: len(x) // 2] + np.arange(len(x) // 2)

    for shape in [(10,), (2, 10), (2, 2, 10)]:
        x = np.ones(shape)
        y = f(x)
        xshape = x.shape
        yshape = y.shape
        assert len(xshape) == len(yshape)
        assert xshape[-1] // 2 == yshape[-1]
        y = f(x=x)
        yshape = y.shape
        assert len(xshape) == len(yshape)
        assert xshape[-1] // 2 == yshape[-1]

    # manually expand 1-loop
    x = np.ones((2, 10), dtype=np.float64)
    y = np.empty((2, 5), dtype=np.float64)
    for i in range(len(x)):
        y[i] = f(x[i])
    yhat = f(x)
    assert np.allclose(yhat, y)

    # expand 2-loop
    x = np.ones((2, 2, 10), dtype=np.float64)
    y = np.empty((2, 2, 5), dtype=np.float64)
    for i in range(len(x)):
        for j in range(len(x[i])):
            y[i][j] = f(x[i][j])
    yhat = f(x)
    assert np.allclose(yhat, y)


def test_multiple_decorators():
    @apply_along_last_axis
    @automatic_type_conversion
    def half_vec(x):
        assert x.ndim == 1
        return x[: len(x) // 2]

    for shape in [(10,), (2, 10), (2, 2, 10)]:
        for dtype in [np.float32, np.float16, np.float64]:
            x = np.ones(shape, dtype=dtype)
            y = half_vec(x)
            xshape = x.shape
            yshape = y.shape
            assert len(xshape) == len(yshape)
            assert xshape[-1] // 2 == yshape[-1]
            assert x.dtype == y.dtype



import numpy as np
import pytest
from pysptk import bartlett, blackman, hamming, hanning, rectangular, trapezoid


@pytest.mark.parametrize(
    "f", [blackman, hanning, hamming, bartlett, trapezoid, rectangular]
)
@pytest.mark.parametrize("n", [16, 128, 256, 1024, 2048, 4096])
def test_windows(f, n):
    def __test(f, N, normalize):
        w = f(N, normalize)
        assert np.all(np.isfinite(w))

    __test(f, n, 1)


def test_windows_corner_case():
    def __test(f, N, normalize):
        f(N, normalize)

    # unsupported normalize flags
    for f in [blackman, hanning, hamming, bartlett, trapezoid, rectangular]:
        with pytest.raises(ValueError):
            __test(f, 256, -1)
        with pytest.raises(ValueError):
            __test(f, 256, 3)



# coding: utf-8
# cython: boundscheck=True, wraparound=True

import numpy as np
cimport numpy as np

cimport cython
from . cimport _sptk

from warnings import warn
from pysptk.util import assert_gamma, assert_fftlen, assert_pade, assert_stage


### Library routines ###

def agexp(r, x, y):
    return _agexp(r, x, y)


def gexp(r, x):
    return _gexp(r, x)


def glog(r, x):
    return _glog(r, x)


def mseq():
    return _mseq()


def acorr(np.ndarray[np.float64_t, ndim=1, mode="c"] x not None, order):
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] r
    r = np.zeros(order + 1)
    _acorr(&x[0], len(x), &r[0], order)
    return r


### Adaptive mel-generalized cepstrum analysis ###

def acep(x, np.ndarray[np.float64_t, ndim=1, mode="c"] c not None,
         lambda_coef=0.98, step=0.1, tau=0.9, pd=4, eps=1.0e-6):
    assert_pade(pd)
    cdef int order = len(c) - 1
    cdef double prederr
    prederr = _acep(x, &c[0], order, lambda_coef, step, tau, pd, eps)
    return prederr


def agcep(x, np.ndarray[np.float64_t, ndim=1, mode="c"] c not None,
          stage=1,
          lambda_coef=0.98, step=0.1, tau=0.9, eps=1.0e-6):
    assert_stage(stage)

    cdef int order = len(c) - 1
    cdef double prederr
    prederr = _agcep(x, &c[0], order, stage, lambda_coef, step, tau, eps)
    return prederr


def amcep(x, np.ndarray[np.float64_t, ndim=1, mode="c"] b not None,
          alpha=0.35,
          lambda_coef=0.98, step=0.1, tau=0.9, pd=4, eps=1.0e-6):
    assert_pade(pd)
    cdef int order = len(b) - 1
    cdef double prederr
    prederr = _amcep(x, &b[0], order, alpha, lambda_coef, step, tau, pd, eps)
    return prederr


### Mel-generalized cepstrum analysis ###

def mcep(np.ndarray[np.float64_t, ndim=1, mode="c"] windowed not None,
         order=25, alpha=0.35,
         miniter=2,
         maxiter=30,
         threshold=0.001,
         etype=0,
         eps=0.0,
         min_det=1.0e-6,
         itype=0):
    if not itype in range(0, 5):
        raise ValueError("unsupported itype: %d, must be in 0:4" % itype)

    if not etype in range(0, 3):
        raise ValueError("unsupported etype: %d, must be in 0:2" % etype)

    if etype == 0 and eps != 0.0:
        raise ValueError("eps cannot be specified for etype = 0")

    if etype == 1 and eps < 0.0:
        raise ValueError("eps: %f, must be >= 0" % eps)

    if etype == 2 and eps >= 0.0:
        raise ValueError("eps: %f, must be < 0" % eps)

    if min_det < 0.0:
        raise ValueError("min_det must be positive: min_det = %f" % min_det)

    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] x
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] mc
    cdef int frame_length
    if itype == 0:
       frame_length = len(windowed)
    else:
       frame_length = (len(windowed) - 1) * 2  # fftlen

    cdef int ret
    mc = np.empty(order + 1, dtype=np.float64)
    x = np.zeros(frame_length, dtype=np.float64)
    x[:len(windowed)] = windowed
    ret = _mcep(&x[0], frame_length, &mc[0],
                order, alpha, miniter, maxiter, threshold, etype, eps,
                min_det, itype)
    assert ret == -1 or ret == 0 or ret == 3 or ret == 4
    if ret == 3:
        raise RuntimeError("failed to compute mcep; error occured in theq")
    elif ret == 4:
        raise RuntimeError(
            "zero(s) are found in periodogram, use eps option to floor")

    return mc


def gcep(np.ndarray[np.float64_t, ndim=1, mode="c"] windowed not None,
         order=25, gamma=0.0,
         miniter=2,
         maxiter=30,
         threshold=0.001,
         etype=0,
         eps=0.0,
         min_det=1.0e-6,
         itype=0,
         norm=False):
    assert_gamma(gamma)
    if not itype in range(0, 5):
        raise ValueError("unsupported itype: %d, must be in 0:4" % itype)

    if not etype in range(0, 3):
        raise ValueError("unsupported etype: %d, must be in 0:2" % etype)

    if etype == 0 and eps != 0.0:
        raise ValueError("eps cannot be specified for etype = 0")

    if (etype == 1 or etype == 2) and eps < 0.0:
        raise ValueError("eps: %f, must be >= 0" % eps)

    if min_det < 0.0:
        raise ValueError("min_det must be positive: min_det = %f" % min_det)

    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] gc
    cdef int windowed_length = len(windowed)
    cdef int ret
    gc = np.empty(order + 1, dtype=np.float64)
    ret = _gcep(&windowed[0], windowed_length, &gc[0], order,
                gamma, miniter, maxiter, threshold, etype, eps, min_det, itype)
    assert ret == -1 or ret == 0 or ret == 3
    if ret == 3:
        raise RuntimeError("failed to compute gcep; error occured in theq")

    if not norm:
        _ignorm(&gc[0], &gc[0], order, gamma)

    return gc


@cython.boundscheck(False)
@cython.wraparound(False)
def mgcep(np.ndarray[np.float64_t, ndim=1, mode="c"] windowed not None,
          order=25, alpha=0.35, gamma=0.0,
          num_recursions=None,
          miniter=2,
          maxiter=30,
          threshold=0.001,
          etype=0,
          eps=0.0,
          min_det=1.0e-6,
          itype=0,
          otype=0):
    assert_gamma(gamma)
    if not itype in range(0, 5):
        raise ValueError("unsupported itype: %d, must be in 0:4" % itype)

    if not etype in range(0, 3):
        raise ValueError("unsupported etype: %d, must be in 0:2" % etype)

    if etype == 0 and eps != 0.0:
        raise ValueError("eps cannot be specified for etype = 0")

    if (etype == 1 or etype == 2) and eps < 0.0:
        raise ValueError("eps: %f, must be >= 0" % eps)

    if min_det < 0.0:
        raise ValueError("min_det must be positive: min_det = %f" % min_det)

    if not otype in range(0, 6):
        raise ValueError("unsupported otype: %d, must be in 0:5" % otype)

    if num_recursions is None:
        num_recursions = len(windowed) - 1

    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] mgc
    cdef int windowed_length = len(windowed)
    cdef int ret
    mgc = np.empty(order + 1, dtype=np.float64)
    ret = _mgcep(&windowed[0], windowed_length, &mgc[0],
                 order, alpha, gamma, num_recursions, miniter, maxiter,
                 threshold, etype, eps, min_det, itype)
    assert ret == -1 or ret == 0 or ret == 3
    if ret == 3:
        raise RuntimeError("failed to compute mgcep; error occured in theq")

    if otype == 0 or otype == 1 or otype == 2 or otype == 4:
        _ignorm(&mgc[0], &mgc[0], order, gamma)

    if otype == 0 or otype == 2 or otype == 4:
        _b2mc(&mgc[0], &mgc[0], order, alpha)

    if otype == 2 or otype == 4:
        _gnorm(&mgc[0], &mgc[0], order, gamma)

    cdef int i = 0
    cdef double g = gamma
    if otype == 4 or otype == 5:
        for i in range(1, len(mgc)):
            mgc[i] *= g

    return mgc


def uels(np.ndarray[np.float64_t, ndim=1, mode="c"] windowed not None,
         order=25,
         miniter=2,
         maxiter=30,
         threshold=0.001,
         etype=0,
         eps=0.0,
         itype=0):
    if not itype in range(0, 5):
        raise ValueError("unsupported itype: %d, must be in 0:4" % itype)

    if not etype in range(0, 3):
        raise ValueError("unsupported etype: %d, must be in 0:2" % etype)

    if etype == 0 and eps != 0.0:
        raise ValueError("eps cannot be specified for etype = 0")

    if (etype == 1 or etype == 2) and eps < 0.0:
        raise ValueError("eps: %f, must be >= 0" % eps)

    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] c
    cdef int windowed_length = len(windowed)
    cdef int ret
    c = np.empty(order + 1, dtype=np.float64)
    ret = _uels(&windowed[0], windowed_length, &c[0], order,
                miniter, maxiter, threshold, etype, eps, itype)
    assert ret == -1 or ret == 0 or ret == 3
    if ret == 3:
        raise RuntimeError(
            "zero(s) are found in periodogram, use eps option to floor")

    return c


def fftcep(np.ndarray[np.float64_t, ndim=1, mode="c"] logsp not None,
           order=25,
           num_iter=0,
           acceleration_factor=0.0):
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] c
    cdef int logsp_length = len(logsp)
    c = np.empty(order + 1, dtype=np.float64)
    _fftcep(&logsp[0], logsp_length, &c[0], order,
            num_iter, acceleration_factor)

    return c


def lpc(np.ndarray[np.float64_t, ndim=1, mode="c"] windowed not None,
        order=25,
        min_det=1.0e-6):
    if min_det < 0.0:
        raise ValueError("min_det must be positive: min_det = %f" % min_det)

    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] a
    cdef int windowed_length = len(windowed)
    cdef int ret
    a = np.empty(order + 1, dtype=np.float64)
    ret = _lpc(&windowed[0], windowed_length, &a[0], order, min_det)
    assert ret == -2 or ret == -1 or ret == 0
    if ret == -2:
        warn("failed to compute `stable` LPC. Please try again with different paramters")
    elif ret == -1:
        raise RuntimeError(
            "failed to compute LPC. Please try again with different parameters")

    return a


### MFCC ###

def mfcc(np.ndarray[np.float64_t, ndim=1, mode="c"] x not None,
         order=14, fs=16000, alpha=0.97, eps=1.0, window_len=None,
         frame_len=None, num_filterbanks=20, cepslift=22, use_dft=False,
         use_hamming=False, czero=False, power=False):
    if not (num_filterbanks > order):
        raise ValueError(
            "Number of filterbanks must be greater than order of MFCC")

    if window_len is None:
        window_len = len(x)
    if frame_len is None:
        frame_len = len(x)

    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] cc
    cc = np.zeros(order + 2)

    cdef Boolean _dft_mode = TR if use_dft else FA
    cdef Boolean _use_hamming = TR if use_hamming else FA

    # after ccall we get
    # mfcc[0], mfcc[1], mfcc[2], ... mfcc[m-1], c0, Power
    _mfcc(&x[0], &cc[0], fs, alpha, eps, window_len, frame_len, order+1,
          num_filterbanks, cepslift, _dft_mode, _use_hamming)

    if (not czero) and power:
        cc[-2] = cc[-1]
    if not power:
        cc = cc[:-1]
    if not czero:
        cc = cc[:-1]

    return cc


### LPC, LSP and PARCOR conversions ###

def lpc2c(np.ndarray[np.float64_t, ndim=1, mode="c"] lpc not None,
          order=None):
    if order is None:
        order = len(lpc) - 1

    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] ceps
    cdef int src_order = len(lpc) - 1
    ceps = np.empty(order + 1, dtype=np.float64)
    _lpc2c(&lpc[0], src_order, &ceps[0], order)
    return ceps


def lpc2lsp(np.ndarray[np.float64_t, ndim=1, mode="c"] lpc not None,
            numsp=128, maxiter=4, eps=1.0e-6, has_gain=True,
            loggain=False, otype=0, fs=None):
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] lsp
    cdef int lpc_start_idx = 1 if has_gain else 0
    cdef int order = len(lpc) - 1 if has_gain else len(lpc)
    cdef int ret
    lsp = np.zeros_like(lpc)
    ret = _lpc2lsp(&lpc[0], &lsp[lpc_start_idx], order, numsp, maxiter, eps)
    if otype == 0:
        lsp[lpc_start_idx:] *= 2 * np.pi
    elif otype == 2 or otype == 3:
        if fs is None:
            raise ValueError("fs must be specified when otype == 2 or 3")
        lsp[lpc_start_idx:] *= fs
    if ret == -1:
      raise RuntimeError("Failed to transform linear predictive coefficients to line spectral pairs")

    if otype == 3:
        lsp[lpc_start_idx:] *= 1000.0

    if has_gain:
      lsp[0] = lpc[0]
      if loggain:
        lsp[0] = np.log(lpc[0])

    return lsp


def lpc2par(np.ndarray[np.float64_t, ndim=1, mode="c"] lpc not None):
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] par
    par = np.empty_like(lpc)
    cdef int order = len(lpc) - 1
    _lpc2par(&lpc[0], &par[0], order)
    return par


def par2lpc(np.ndarray[np.float64_t, ndim=1, mode="c"] par not None):
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] lpc
    lpc = np.empty_like(par)
    cdef int order = len(par) - 1
    _par2lpc(&par[0], &lpc[0], order)
    return lpc


def lsp2lpc(np.ndarray[np.float64_t, ndim=1, mode="c"] lsp not None,
    has_gain=True, loggain=False, fs=None, itype=0):
    lsp = lsp.copy()
    cdef int lpc_start_idx = 1 if has_gain else 0

    if loggain and not has_gain:
        raise ValueError("has_gain must be True if you set loggain=True")

    if itype == 0:
        lsp[lpc_start_idx:] /= 2 * np.pi
    elif itype == 2 or itype == 3:
        if fs is None:
            raise ValueError("fs must be specified when itype == 2 or 3")
        lsp[lpc_start_idx:] /= fs

    if itype == 3:
        lsp[lpc_start_idx:] /= 1000.0

    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] lpc
    cdef int order
    if has_gain:
      order = len(lsp) - 1
    else:
      order = len(lsp)
    lpc = np.empty_like(lsp)

    _lsp2lpc(&lsp[lpc_start_idx], &lpc[0], order)

    if has_gain:
      lpc[0] = lsp[0]
      if loggain:
        lpc[0] = np.exp(lpc[0])

    return lpc


def lsp2sp(np.ndarray[np.float64_t, ndim=1, mode="c"] lsp not None,
           fftlen=256, has_gain=True, loggain=False, fs=None, itype=0):
    assert_fftlen(fftlen)
    lsp = lsp.copy()
    cdef int lsp_start_idx = 1 if has_gain else 0

    if itype == 1:
        lsp[lsp_start_idx:] *= 2 * np.pi
    elif itype == 2 or itype == 3:
        if fs is None:
            raise ValueError("fs must be specified when itype == 2 or 3")
        lsp[lsp_start_idx:] = lsp[lsp_start_idx:] / fs * 2 * np.pi

    if itype == 3:
        lsp[lsp_start_idx:] /= 1000.0

    if loggain:
        lsp[0] = np.log(lsp[0])

    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] sp
    cdef int sp_length = (fftlen >> 1) + 1
    sp = np.empty(sp_length, dtype=np.float64)
    cdef int order = len(lsp) - 1
    _lsp2sp(&lsp[0], order, &sp[0], sp_length, 1 if has_gain else 0)
    return sp


### Mel-generalized cepstrum conversions ###

def mc2b(np.ndarray[np.float64_t, ndim=1, mode="c"] mc not None,
         alpha=0.35):
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] b
    b = np.empty_like(mc)
    cdef int order = len(mc) - 1
    _mc2b(&mc[0], &b[0], order, alpha)
    return b


def b2mc(np.ndarray[np.float64_t, ndim=1, mode="c"] b not None,
         alpha=0.35):
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] mc
    mc = np.empty_like(b)
    cdef int order = len(b) - 1
    _b2mc(&b[0], &mc[0], order, alpha)
    return mc


def b2c(np.ndarray[np.float64_t, ndim=1, mode="c"] b not None,
        dst_order=None,
        alpha=0.35):
    cdef int src_order = len(b) - 1
    if dst_order is None:
        dst_order = src_order
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] c
    c = np.empty(dst_order + 1, dtype=np.float64)
    _b2c(&b[0], src_order, &c[0], dst_order, alpha)
    return c


def c2acr(np.ndarray[np.float64_t, ndim=1, mode="c"] c not None,
          order=None,
          fftlen=256):
    assert_fftlen(fftlen)
    if len(c) > fftlen:
        raise ValueError("fftlen must be larger than length of input")
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] r
    cdef int src_order = len(c) - 1
    if order is None:
        order = src_order
    r = np.empty(order + 1, dtype=np.float64)
    _c2acr(&c[0], src_order, &r[0], order, fftlen)
    return r


def c2ir(np.ndarray[np.float64_t, ndim=1, mode="c"] c not None,
         length=256):
    cdef int order = len(c)  # NOT len(c) - 1
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] h
    h = np.empty(length, dtype=np.float64)
    _c2ir(&c[0], order, &h[0], length)
    return h


def ic2ir(np.ndarray[np.float64_t, ndim=1, mode="c"] h not None,
          order=25):
    cdef int length = len(h)
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] c
    c = np.empty(order + 1, dtype=np.float64)
    _ic2ir(&h[0], length, &c[0], len(c))
    return c


@cython.boundscheck(False)
@cython.wraparound(False)
def c2ndps(np.ndarray[np.float64_t, ndim=1, mode="c"] c not None,
           fftlen=256):
    assert_fftlen(fftlen)
    cdef int dst_length = (fftlen >> 1) + 1
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] ndps, buf
    ndps = np.empty(dst_length, dtype=np.float64)
    cdef int order = len(c) - 1
    buf = np.empty(fftlen, dtype=np.float64)
    _c2ndps(&c[0], order, &buf[0], fftlen)

    ndps[:] = buf[0:dst_length]

    return ndps


def ndps2c(np.ndarray[np.float64_t, ndim=1, mode="c"] ndps not None,
           order=25):
    # assuming the lenght of ndps is fftlen/2+1
    cdef int fftlen = (len(ndps) - 1) << 1
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] c
    assert_fftlen(fftlen)
    c = np.empty(order + 1, dtype=np.float64)
    _ndps2c(&ndps[0], fftlen, &c[0], order)
    return c


def gc2gc(np.ndarray[np.float64_t, ndim=1, mode="c"] src_ceps not None,
          src_gamma=0.0, dst_order=None, dst_gamma=0.0):
    assert_gamma(src_gamma)
    assert_gamma(dst_gamma)

    cdef int src_order = len(src_ceps) - 1
    if dst_order is None:
        dst_order = src_order
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] dst_ceps
    dst_ceps = np.empty(dst_order + 1, dtype=np.float64)

    _gc2gc(&src_ceps[0], src_order, src_gamma,
           &dst_ceps[0], dst_order, dst_gamma)

    return dst_ceps


def gnorm(np.ndarray[np.float64_t, ndim=1, mode="c"] ceps not None,
          gamma=0.0):
    assert_gamma(gamma)
    cdef int order = len(ceps) - 1
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] dst_ceps
    dst_ceps = np.empty_like(ceps)
    _gnorm(&ceps[0], &dst_ceps[0], order, gamma)
    return dst_ceps


def ignorm(np.ndarray[np.float64_t, ndim=1, mode="c"] ceps not None,
           gamma=0.0):
    assert_gamma(gamma)
    cdef int order = len(ceps) - 1
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] dst_ceps
    dst_ceps = np.empty_like(ceps)
    _ignorm(&ceps[0], &dst_ceps[0], order, gamma)
    return dst_ceps


def freqt(np.ndarray[np.float64_t, ndim=1, mode="c"] ceps not None,
          order=25, alpha=0.0):
    cdef int src_order = len(ceps) - 1
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] dst_ceps
    dst_ceps = np.empty(order + 1, dtype=np.float64)
    _freqt(&ceps[0], src_order, &dst_ceps[0], order, alpha)
    return dst_ceps


def frqtr(np.ndarray[np.float64_t, ndim=1, mode="c"] src_ceps not None,
          order=25, alpha=0.0):
    cdef int src_order = len(src_ceps) - 1
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] dst_ceps
    dst_ceps = np.empty(order + 1, dtype=np.float64)
    _frqtr(&src_ceps[0], src_order, &dst_ceps[0], order, alpha)
    return dst_ceps


def mgc2mgc(np.ndarray[np.float64_t, ndim=1, mode="c"] src_ceps not None,
            src_alpha=0.0, src_gamma=0.0,
            dst_order=None, dst_alpha=0.0, dst_gamma=0.0):
    assert_gamma(src_gamma)
    assert_gamma(dst_gamma)

    cdef int src_order = len(src_ceps) - 1
    if dst_order is None:
        dst_order = src_order
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] dst_ceps
    dst_ceps = np.empty(dst_order + 1, dtype=np.float64)

    _mgc2mgc(&src_ceps[0], src_order, src_alpha, src_gamma,
             &dst_ceps[0], dst_order, dst_alpha, dst_gamma)

    return dst_ceps


@cython.boundscheck(False)
@cython.wraparound(False)
def mgc2sp(np.ndarray[np.float64_t, ndim=1, mode="c"] ceps not None,
           alpha=0.0, gamma=0.0, fftlen=256):
    assert_gamma(gamma)
    assert_fftlen(fftlen)

    cdef int order = len(ceps) - 1
    cdef np.ndarray[np.complex128_t, ndim = 1, mode = "c"] sp
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] sp_r, sp_i

    sp = np.empty((fftlen >> 1) + 1, dtype=np.complex128)
    sp_r = np.zeros(fftlen, dtype=np.float64)
    sp_i = np.zeros(fftlen, dtype=np.float64)

    _mgc2sp(&ceps[0], order, alpha, gamma, &sp_r[0], &sp_i[0], fftlen)

    cdef int i
    for i in range(0, len(sp)):
        sp[i] = sp_r[i] + sp_i[i] * 1j

    return sp


def mgclsp2sp(np.ndarray[np.float64_t, ndim=1, mode="c"] lsp not None,
              alpha=0.0, gamma=0.0, fftlen=256, gain=True):
    assert_gamma(gamma)
    assert_fftlen(fftlen)

    cdef int order = gain if len(lsp) - 1 else len(lsp)
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] sp
    sp = np.empty((fftlen >> 1) + 1, dtype=np.float64)

    _mgclsp2sp(alpha, gamma, &lsp[0], order, &sp[0], len(sp), int(gain))

    return sp


### F0 analysis ###

def swipe(np.ndarray[np.float64_t, ndim=1, mode="c"] x not None,
          fs, hopsize,
          min=60.0, max=240.0, threshold=0.3, otype="f0"):
    supported_otypes = ["pitch", "f0", "logf0"]
    if isinstance(otype, int) and (not otype in range(0, 3)) or \
        isinstance(otype, str) and not otype in supported_otypes:
        raise ValueError("otype must be (0) pitch, (1) f0, or (2) log(f0)")

    if isinstance(otype, str):
        otype = supported_otypes.index(otype)

    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] f0
    cdef int x_length = len(x)
    cdef int expected_len = int(np.ceil(float(x_length) / hopsize))

    f0 = np.empty(expected_len, dtype=np.float64)

    _swipe(&x[0], &f0[0], x_length, fs, hopsize, min, max, threshold, otype)
    return f0


def rapt(np.ndarray[np.float32_t, ndim=1, mode="c"] x not None,
         fs, hopsize,
         min=60, max=240, voice_bias=0.0, otype="f0"):
    supported_otypes = ["pitch", "f0", "logf0"]
    if isinstance(otype, int) and (not otype in range(0, 3)) or \
       isinstance(otype, str) and not otype in supported_otypes:
        raise ValueError("otype must be (0) pitch, (1) f0, or (2) log(f0) ")

    if isinstance(otype, str):
        otype = supported_otypes.index(otype)

    if min >=max or max >= fs//2 or min <= float(fs)/10000.0:
        raise ValueError("invalid min/max frequency parameters")

    frame_period = float(hopsize) / fs
    frame_period = float(int(0.5 + (fs * frame_period))) / fs
    if frame_period > 0.1 or frame_period < 1.0/fs:
       raise ValueError("frame period must be between [1/fs, 0.1]")

    cdef np.ndarray[np.float32_t, ndim = 1, mode = "c"] f0
    cdef int x_length = len(x)
    cdef int expected_len = int(np.ceil(float(x_length) / hopsize))
    cdef int ret

    f0 = np.empty(expected_len, dtype=np.float32)

    ret = _rapt(&x[0], &f0[0], x_length, fs, hopsize, min, max,
                voice_bias, otype)
    if ret == 2:
        raise ValueError("input range too small for analysis by get_f0")
    elif ret == 3:
        raise RuntimeError("problem in init_dp_f0()")

    assert ret == 0

    return f0


### Window functions ###

cdef __window(Window window_type, np.ndarray[np.float64_t, ndim=1, mode="c"] x,
              int size, int normalize):
    if normalize < 0 or normalize > 2:
        raise ValueError("normalize must be 0, 1 or 2")
    cdef double g = _window(window_type, &x[0], size, normalize)
    return x


def blackman(n, normalize=1):
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] x
    x = np.ones(n, dtype=np.float64)
    cdef Window window_type = BLACKMAN
    return __window(window_type, x, len(x), normalize)


def hamming(n, normalize=1):
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] x
    x = np.ones(n, dtype=np.float64)
    cdef Window window_type = HAMMING
    return __window(window_type, x, len(x), normalize)


def hanning(n, normalize=1):
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] x
    x = np.ones(n, dtype=np.float64)
    cdef Window window_type = HANNING
    return __window(window_type, x, len(x), normalize)


def bartlett(n, normalize=1):
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] x
    x = np.ones(n, dtype=np.float64)
    cdef Window window_type = BARTLETT
    return __window(window_type, x, len(x), normalize)


def trapezoid(n, normalize=1):
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] x
    x = np.ones(n, dtype=np.float64)
    cdef Window window_type = TRAPEZOID
    return __window(window_type, x, len(x), normalize)


def rectangular(n, normalize=1):
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] x
    x = np.ones(n, dtype=np.float64)
    cdef Window window_type = RECTANGULAR
    return __window(window_type, x, len(x), normalize)


### Waveform generation filters ###

def zerodf_delay_length(int order):
    return order

def zerodf(x, np.ndarray[np.float64_t, ndim=1, mode="c"] a not None,
           np.ndarray[np.float64_t, ndim=1, mode="c"] delay not None):
    cdef int order = len(a) - 1
    if len(delay) != zerodf_delay_length(order):
        raise ValueError("inconsistent delay length")

    return _zerodf(x, &a[0], order, &delay[0])

def zerodft(x, np.ndarray[np.float64_t, ndim=1, mode="c"] a not None,
            np.ndarray[np.float64_t, ndim=1, mode="c"] delay not None):
    cdef int order = len(a) - 1
    if len(delay) != zerodf_delay_length(order):
        raise ValueError("inconsistent delay length")

    return _zerodft(x, &a[0], order, &delay[0])

def poledf_delay_length(int order):
    return order

def poledf(x, np.ndarray[np.float64_t, ndim=1, mode="c"] a not None,
           np.ndarray[np.float64_t, ndim=1, mode="c"] delay not None):
    cdef int order = len(a) - 1
    if len(delay) != poledf_delay_length(order):
        raise ValueError("inconsistent delay length")

    return _poledf(x, &a[0], order, &delay[0])

def poledft(x, np.ndarray[np.float64_t, ndim=1, mode="c"] a not None,
            np.ndarray[np.float64_t, ndim=1, mode="c"] delay not None):
    cdef int order = len(a) - 1
    if len(delay) != poledf_delay_length(order):
        raise ValueError("inconsistent delay length")

    return _poledft(x, &a[0], order, &delay[0])

def lmadf_delay_length(int order, int pd):
    return 2 * pd * (order + 1)

def lmadf(x, np.ndarray[np.float64_t, ndim=1, mode="c"] b not None,
          pd,
          np.ndarray[np.float64_t, ndim=1, mode="c"] delay not None):
    assert_pade(pd)

    cdef int order = len(b) - 1
    if len(delay) != lmadf_delay_length(order, pd):
        raise ValueError("inconsistent delay length")

    return _lmadf(x, &b[0], order, pd, &delay[0])

def lspdf_delay_length(int order):
    return 2 * order + 1

def lspdf(x, np.ndarray[np.float64_t, ndim=1, mode="c"] f not None,
          np.ndarray[np.float64_t, ndim=1, mode="c"] delay not None):
    cdef int order = len(f) - 1
    if len(delay) != lspdf_delay_length(order):
        raise ValueError("inconsistent delay length")

    if order % 2 == 0:
        return _lspdf_even(x, &f[0], order, &delay[0])
    else:
        return _lspdf_odd(x, &f[0], order, &delay[0])

def ltcdf_delay_length(int order):
    return order + 1

def ltcdf(x, np.ndarray[np.float64_t, ndim=1, mode="c"] k not None,
          np.ndarray[np.float64_t, ndim=1, mode="c"] delay not None):
    cdef int order = len(k) - 1
    if len(delay) != ltcdf_delay_length(order):
        raise ValueError("inconsistent delay length")

    return _ltcdf(x, &k[0], order, &delay[0])

def glsadf_delay_length(int order, int stage):
    return order * (stage + 1) + 1

def glsadf(x, np.ndarray[np.float64_t, ndim=1, mode="c"] c not None,
           stage,
           np.ndarray[np.float64_t, ndim=1, mode="c"] delay not None):
    assert_stage(stage)

    cdef int order = len(c) - 1
    if len(delay) != glsadf_delay_length(order, stage):
        raise ValueError("inconsistent delay length")

    return _glsadf(x, &c[0], order, stage, &delay[0])

def glsadft(x, np.ndarray[np.float64_t, ndim=1, mode="c"] c not None,
            stage,
            np.ndarray[np.float64_t, ndim=1, mode="c"] delay not None):
    assert_stage(stage)

    cdef int order = len(c) - 1
    if len(delay) != glsadf_delay_length(order, stage):
        raise ValueError("inconsistent delay length")

    return _glsadft(x, &c[0], order, stage, &delay[0])

def mlsadf_delay_length(int order, int pd):
    return 3 * (pd + 1) + pd * (order + 2)

def mlsadf(x, np.ndarray[np.float64_t, ndim=1, mode="c"] b not None,
           alpha, pd,
           np.ndarray[np.float64_t, ndim=1, mode="c"] delay not None):
    assert_pade(pd)

    cdef int order = len(b) - 1
    if len(delay) != mlsadf_delay_length(order, pd):
        raise ValueError("inconsistent delay length")

    return _mlsadf(x, &b[0], order, alpha, pd, &delay[0])

def mlsadft(x, np.ndarray[np.float64_t, ndim=1, mode="c"] b not None,
            alpha, pd,
            np.ndarray[np.float64_t, ndim=1, mode="c"] delay not None):
    assert_pade(pd)

    cdef int order = len(b) - 1
    if len(delay) != mlsadf_delay_length(order, pd):
        raise ValueError("inconsistent delay length")

    return _mlsadft(x, &b[0], order, alpha, pd, &delay[0])

def mglsadf_delay_length(int order, int stage):
    return (order + 1) * stage

def mglsadf(x, np.ndarray[np.float64_t, ndim=1, mode="c"] b not None,
            alpha, stage,
            np.ndarray[np.float64_t, ndim=1, mode="c"] delay not None):
    assert_stage(stage)

    cdef int order = len(b) - 1
    if len(delay) != mglsadf_delay_length(order, stage):
        raise ValueError("inconsistent delay length")

    return _mglsadf(x, &b[0], order, alpha, stage, &delay[0])

def mglsadft(x, np.ndarray[np.float64_t, ndim=1, mode="c"] b not None,
             alpha, stage,
             np.ndarray[np.float64_t, ndim=1, mode="c"] delay not None):
    assert_stage(stage)

    cdef int order = len(b) - 1
    if len(delay) != mglsadf_delay_length(order, stage):
        raise ValueError("inconsistent delay length")

    return _mglsadft(x, &b[0], order, alpha, stage, &delay[0])

### Excitation ###

def excite(np.ndarray[np.float64_t, ndim=1, mode = "c"] pitch, frame_period=100, interp_period=1, gaussian=False, seed=1):
    # Allocate memory for output
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] excitation
    cdef int n = len(pitch)
    cdef int expected_len = int(frame_period*(n-1))

    excitation = np.empty(expected_len, dtype=np.float64)
    # Call
    _excite(&pitch[0], n, &excitation[0], frame_period, interp_period, gaussian, seed)
    # Return allocated output
    return excitation


### Utils ###

def phidf(x, order, alpha,
          np.ndarray[np.float64_t, ndim=1, mode="c"] delay not None):
    if len(delay) != order + 1:
        raise ValueError("inconsistent order or delay")

    _phidf(x, order, alpha, &delay[0])


def lspcheck(np.ndarray[np.float64_t, ndim=1, mode="c"] lsp not None):
    cdef int ret = _lspcheck(&lsp[0], len(lsp) - 1)
    return ret


def levdur(np.ndarray[np.float64_t, ndim=1, mode="c"] r not None, double eps):
    cdef int order = len(r) - 1
    cdef np.ndarray[np.float64_t, ndim = 1, mode = "c"] a
    a = np.empty(order + 1, dtype=np.float64)
    cdef int ret = _levdur(&r[0], &a[0], order, eps)
    if ret == -1:
      raise RuntimeError("abnormally completed")
    elif ret == -2:
      raise RuntimeError("Unstable LPC")

    return a



"""
Library routines
----------------
.. autosummary::
    :toctree: generated/

    agexp
    gexp
    glog
    mseq
    acorr

Adaptive cepstrum analysis
--------------------------
.. autosummary::
    :toctree: generated/

    acep
    agcep
    amcep

Mel-generalized cepstrum analysis
---------------------------------
.. autosummary::
    :toctree: generated/

    mcep
    gcep
    mgcep
    uels
    fftcep
    lpc

MFCC
----
.. autosummary::
    :toctree: generated/

    mfcc

LPC, LSP and PARCOR conversions
-------------------------------
.. autosummary::
    :toctree: generated/

    lpc2c
    lpc2lsp
    lpc2par
    par2lpc
    lsp2lpc
    lsp2sp

Mel-generalized cepstrum conversions
------------------------------------
.. autosummary::
    :toctree: generated/

    mc2b
    b2mc
    c2acr
    levdur
    c2ir
    ic2ir
    c2ndps
    ndps2c
    gc2gc
    gnorm
    ignorm
    freqt
    mgc2mgc
    mgc2sp
    mgclsp2sp

F0 analysis
-----------
.. autosummary::
    :toctree: generated/

    swipe
    rapt

Excitation generation
---------------------
.. autosummary::
    :toctree: generated/

    excite

Window functions
----------------
.. autosummary::
    :toctree: generated/

    blackman
    hamming
    hanning
    bartlett
    trapezoid
    rectangular

Waveform generation filters
---------------------------
.. autosummary::
    :toctree: generated/

    zerodf
    zerodft
    poledf
    poledft
    lmadf
    lspdf
    ltcdf
    glsadf
    glsadft
    mlsadf
    mlsadft
    mglsadf
    mglsadft

Utilities for waveform generation filters
-----------------------------------------
.. autosummary::
    :toctree: generated/

    zerodf_delay
    poledf_delay
    lmadf_delay
    lspdf_delay
    ltcdf_delay
    glsadf_delay
    mlsadf_delay
    mglsadf_delay

Metrics
-------
.. autosummary::
    :toctree: generated/

    cdist

"""

import numpy as np
from pysptk.util import (
    apply_along_last_axis,
    automatic_type_conversion,
    automatic_type_conversion_float32,
)

from . import _sptk  # type: ignore


def cdist(c1, c2, otype=0, frame=False):
    """Calculation of cepstral distance

    Parameters
    ----------
    c1 : array
        Minimum-phase cepstrum
    c2 : array
        Minimum-phase cepstrum
    otype : int
        Output data type
            (0) [db]
            (1) squared error
            (2) root squared error

        Default is 0.
    frame : bool
        If True, returns frame-wise distance, otherwise returns mean distance.
        Default is False.

    Returns
    -------
    distance

    """

    if otype not in [0, 1, 2]:
        raise ValueError("unsupported otype: %d, must be in 0:2" % otype)

    assert c1.shape[0] == c2.shape[0]
    c1.shape[0]

    s = ((c1[:, 1:] - c2[:, 1:]) ** 2).sum(-1)

    if otype == 0:
        s = np.sqrt(2 * s) * 10 / np.log(10)
    elif otype == 2:
        s = np.sqrt(s)

    if frame:
        return s
    else:
        return s.mean()


# Library routines #


def agexp(r, x, y):
    """Magnitude squared generalized exponential function

    Parameters
    ----------
    r : float
        Gamma
    x : float
        Real part
    y : float
        Imaginary part

    Returns
    -------
    Value

    """
    return _sptk.agexp(r, x, y)


def gexp(r, x):
    """Generalized exponential function

    Parameters
    ----------
    r : float
        Gamma

    x : float
        Arg

    Returns
    -------
    Value

    """
    return _sptk.gexp(r, x)


def glog(r, x):
    """Generalized logarithmic function

    Parameters
    ----------
    r : float
        Gamma
    x : float
        Arg

    Returns
    -------
    Value

    """
    return _sptk.glog(r, x)


def mseq():
    """M-sequence

    Returns
    -------
    A sample of m-sequence

    """
    return _sptk.mseq()


@apply_along_last_axis
@automatic_type_conversion
def acorr(x, order):
    """Autocorrelation

    Parameters
    ----------
    x : array
        Input frame

    order : int
        Order of sequence

    Returns
    -------
    r : array
        Autocorrelation

    See Also
    --------
    pysptk.sptk.levdur
    pysptk.sptk.c2acr
    """
    return _sptk.acorr(x, order)


# Adaptive mel-generalized cepstrum analysis #


def acep(x, c, lambda_coef=0.98, step=0.1, tau=0.9, pd=4, eps=1.0e-6):
    """Adaptive cepstral analysis

    Parameters
    ----------
    x : double
        A input sample

    c : array, shape(``order + 1``)
        Cepstrum. The result is stored in place.

    lambda_coef : float, optional
        Leakage factor. Default is 0.98.

    step : float, optional
        Step size. Default is 0.1.

    tau : float, optional
        Momentum constant. Default is 0.9.

    pd : int, optional
        Order of pade approximation. Default is 4.

    eps : float, optional
        Minimum value for epsilon. Default is 1.0e-6.

    Returns
    -------
    prederr : float
        Prediction error

    Raises
    ------
    ValueError
        if invalid order of pade approximation is specified

    See Also
    --------
    pysptk.sptk.uels
    pysptk.sptk.gcep
    pysptk.sptk.mcep
    pysptk.sptk.mgcep
    pysptk.sptk.amcep
    pysptk.sptk.agcep
    pysptk.sptk.lmadf

    """
    return _sptk.acep(x, c, lambda_coef, step, tau, pd, eps)


def agcep(x, c, stage=1, lambda_coef=0.98, step=0.1, tau=0.9, eps=1.0e-6):
    """Adaptive generalized cepstral analysis

    Parameters
    ----------
    x : float
        A input sample

    c : array, shape(``order + 1``), optional
        Cepstrum. The result is stored in-place.

    stage : int, optional
        -1 / gamma. Default is 1.

    lambda_coef : float, optional
        Leakage factor. Default is 0.98.

    step : float, optional
        Step size. Default is 0.1.

    tau : float, optional
        Momentum constant. Default is 0.9.

    eps : float, optional
        Minimum value for epsilon. Default is 1.0e-6.

    Returns
    -------
    prederr : float
        Prediction error

    Raises
    ------
    ValueError
        if invalid number of stage is specified

    See Also
    --------
    pysptk.sptk.acep
    pysptk.sptk.amcep
    pysptk.sptk.glsadf

    """
    return _sptk.agcep(x, c, stage, lambda_coef, step, tau, eps)


def amcep(x, b, alpha=0.35, lambda_coef=0.98, step=0.1, tau=0.9, pd=4, eps=1.0e-6):
    """Adaptive mel-cepstral analysis

    Parameters
    ----------
    x : float
        A input sample

    b : array, shape(``order + 1``), optional
        MLSA filter coefficients. The result is stored in-place.

    alpha : float, optional
        All-pass constant. Default is 0.35.

    lambda_coef : float, optional
        Leakage factor. Default is 0.98.

    step : float, optional
        Step size. Default is 0.1.

    tau : float, optional
        Momentum constant. Default is 0.9.

    pd : int, optional
        Order of pade approximation. Default is 4.

    eps : float, optional
        Minimum value for epsilon. Default is 1.0e-6.

    Returns
    -------
    prederr : float
        Prediction error

    Raises
    ------
    ValueError
        if invalid order of pade approximation is specified

    See Also
    --------
    pysptk.sptk.acep
    pysptk.sptk.agcep
    pysptk.sptk.mc2b
    pysptk.sptk.b2mc
    pysptk.sptk.mlsadf

    """
    return _sptk.amcep(x, b, alpha, lambda_coef, step, tau, pd, eps)


# Mel-generalized cepstrum analysis #


@apply_along_last_axis
@automatic_type_conversion
def mcep(
    windowed,
    order=25,
    alpha=0.35,
    miniter=2,
    maxiter=30,
    threshold=0.001,
    etype=0,
    eps=0.0,
    min_det=1.0e-6,
    itype=0,
):
    """Mel-cepstrum analysis

    Parameters
    ----------
    windowed : array, shape (``frame_len``)
        A windowed frame

    order : int, optional
        Order of mel-cepstrum. Default is 25.

    alpha : float, optional
        All pass constant. Default is 0.35.

    miniter : int, optional
        Minimum number of iteration. Default is 2.

    maxiter : int, optional
        Maximum number of iteration. Default is 30.

    threshold : float, optional
        Threshold in theq. Default is 0.001.

    etype : int, optional
        Type of parameter ``eps``
             (0) not used
             (1) initial value of log-periodogram
             (2) floor of periodogram in db

        Default is 0.

    eps : float, optional
        Initial value for log-periodogram or floor of periodogram in db.
        Default is 0.0.

    min_det : float, optional
        Mimimum value of the determinant of normal matrix.
        Default is 1.0e-6

    itype : float, optional
        Input data type
            (0) windowed signal
            (1) log amplitude in db
            (2) log amplitude
            (3) amplitude
            (4) periodogram

        Default is 0.

    Returns
    -------
    mc : array, shape (``order + 1``)
        Mel-cepstrum

    Raises
    ------
    ValueError
        - if invalid ``itype`` is specified
        - if invalid ``etype`` is specified
        - if nonzero ``eps`` is specified when etype = 0
        - if negative ``eps`` is specified
        - if negative ``min_det`` is specified

    RuntimeError
        - if zero(s) are found in periodogram
        - if error happened in theq

    See Also
    --------
    pysptk.sptk.uels
    pysptk.sptk.gcep
    pysptk.sptk.mgcep
    pysptk.sptk.mlsadf

    """
    return _sptk.mcep(
        windowed, order, alpha, miniter, maxiter, threshold, etype, eps, min_det, itype
    )


@apply_along_last_axis
@automatic_type_conversion
def gcep(
    windowed,
    order=25,
    gamma=0.0,
    miniter=2,
    maxiter=30,
    threshold=0.001,
    etype=0,
    eps=0.0,
    min_det=1.0e-6,
    itype=0,
    norm=False,
):
    """Generalized-cepstrum analysis

    Parameters
    ----------
    windowed : array, shape (``frame_len``)
        A windowed frame

    order : int, optional
        Order of generalized-cepstrum. Default is 25.

    gamma : float, optional
        Parameter of generalized log function. Default is 0.0.

    miniter : int, optional
        Minimum number of iteration. Default is 2.

    maxiter : int, optional
        Maximum number of iteration. Default is 30.

    threshold : float, optional
        Threshold in theq. Default is 0.001

    etype : int, optional
        Type of parameter ``eps``
             (0) not used
             (1) initial value of log-periodogram
             (2) floor of periodogram in db

        Default is 0.

    eps : float, optional
        Initial value for log-periodogram or floor of periodogram in db.
        Default is 0.0.

    min_det : float, optional
        Mimimum value of the determinant of normal matrix. Default is 1.0e-6.

    itype : float, optional
        Input data type
            (0) windowed signal
            (1) log amplitude in db
            (2) log amplitude
            (3) amplitude
            (4) periodogram

        Default is 0.

    Returns
    -------
    gc : array, shape (``order + 1``)
        Generalized cepstrum

    Raises
    ------
    ValueError
        - if invalid ``itype`` is specified
        - if invalid ``etype`` is specified
        - if nonzero ``eps`` is specified when etype = 0
        - if negative ``eps`` is specified
        - if negative ``min_det`` is specified

    RuntimeError
        - if error happened in theq

    See Also
    --------
    pysptk.sptk.uels
    pysptk.sptk.mcep
    pysptk.sptk.mgcep
    pysptk.sptk.glsadf

    """

    return _sptk.gcep(
        windowed, order, gamma, miniter, maxiter, threshold, etype, eps, min_det, itype
    )


@apply_along_last_axis
@automatic_type_conversion
def mgcep(
    windowed,
    order=25,
    alpha=0.35,
    gamma=0.0,
    num_recursions=None,
    miniter=2,
    maxiter=30,
    threshold=0.001,
    etype=0,
    eps=0.0,
    min_det=1.0e-6,
    itype=0,
    otype=0,
):
    """Mel-generalized cepstrum analysis

    Parameters
    ----------
    windowed : array, shape (``frame_len``)
        A windowed frame

    order : int, optional
        Order of mel-generalized cepstrum. Default is 25.

    alpha : float, optional
        All pass constant. Default is 0.35.

    gamma : float, optional
        Parameter of generalized log function. Default is 0.0.

    num_recursions : int, optional
        Number of recursions. Default is ``len(windowed) - 1``.

    miniter : int, optional
        Minimum number of iteration. Default is 2.

    maxiter : int, optional
        Maximum number of iteration. Default is 30.

    threshold : float, optional
        Threshold. Default is 0.001.

    etype : int, optional
        Type of paramter ``e``
             (0) not used
             (1) initial value of log-periodogram
             (2) floor of periodogram in db

        Default is 0.

    eps : float, optional
        Initial value for log-periodogram or floor of periodogram in db.
        Default is 0.0.

    min_det : float, optional
        Mimimum value of the determinant of normal matrix.
        Default is 1.0e-6.

    itype : float, optional
        Input data type
            (0) windowed signal
            (1) log amplitude in db
            (2) log amplitude
            (3) amplitude
            (4) periodogram

        Default is 0.

    otype : int, optional
        Output data type
            (0) mel generalized cepstrum: (c~0...c~m)
            (1) MGLSA filter coefficients: b0...bm
            (2) K~,c~'1...c~'m
            (3) K,b'1...b'm
            (4) K~,g*c~'1...g*c~'m
            (5) K,g*b'1...g*b'm

        Default is 0.

    Returns
    -------
    mgc : array, shape (``order + 1``)
        mel-generalized cepstrum

    Raises
    ------
    ValueError
        - if invalid ``itype`` is specified
        - if invalid ``etype`` is specified
        - if nonzero ``eps`` is specified when etype = 0
        - if negative ``eps`` is specified
        - if negative ``min_det`` is specified
        - if invalid ``otype`` is specified

    RuntimeError
        - if error happened in theq

    See Also
    --------
    pysptk.sptk.uels
    pysptk.sptk.gcep
    pysptk.sptk.mcep
    pysptk.sptk.freqt
    pysptk.sptk.gc2gc
    pysptk.sptk.mgc2mgc
    pysptk.sptk.gnorm
    pysptk.sptk.mglsadf

    """

    return _sptk.mgcep(
        windowed,
        order,
        alpha,
        gamma,
        num_recursions,
        miniter,
        maxiter,
        threshold,
        etype,
        eps,
        min_det,
        itype,
        otype,
    )


@apply_along_last_axis
@automatic_type_conversion
def uels(
    windowed,
    order=25,
    miniter=2,
    maxiter=30,
    threshold=0.001,
    etype=0,
    eps=0.0,
    itype=0,
):
    """Unbiased estimation of log spectrum

    Parameters
    ----------
    windowed : array, shape (``frame_len``)
        A windowed frame

    order : int, optional
        Order of cepstrum. Default is 25.

    miniter : int, optional
        Minimum number of iteration. Default is 2.

    maxiter : int, optional
        Maximum number of iteration. Default is 30.

    threshold : float, optional
        Threshold in theq. Default is 0.001

    etype : int, optional
        Type of parameter ``eps``
             (0) not used
             (1) initial value of log-periodogram
             (2) floor of periodogram in db

        Default is 0.

    eps : float, optional
        Initial value for log-periodogram or floor of periodogram in db.
        Default is 0.0.

    itype : float, optional
        Input data type
            (0) windowed signal
            (1) log amplitude in db
            (2) log amplitude
            (3) amplitude
            (4) periodogram

        Default is 0.

    Returns
    -------
    c : array, shape (``order + 1``)
        cepstrum estimated by uels

    Raises
    ------
    ValueError
        - if invalid ``itype`` is specified
        - if invalid ``etype`` is specified
        - if nonzero ``eps`` is specified when etype = 0
        - if negative ``eps`` is specified

    RuntimeError
        - if zero(s) are found in periodogram

    See Also
    --------
    pysptk.sptk.gcep
    pysptk.sptk.mcep
    pysptk.sptk.mgcep
    pysptk.sptk.lmadf

    """

    return _sptk.uels(windowed, order, miniter, maxiter, threshold, etype, eps, itype)


@apply_along_last_axis
@automatic_type_conversion
def fftcep(logsp, order=25, num_iter=0, acceleration_factor=0.0):
    """FFT-based cepstrum analysis

    Parameters
    ----------
    logsp : array, shape (``frame_len``)
        Log power spectrum

    order : int, optional
        Order of cepstrum. Default is 25.

    num_iter : int, optional
        Number of iteration. Default is 0.

    acceleration_factor : float, optional
        Acceleration factor. Default is 0.0.

    Returns
    -------
    c : array, shape (``order + 1``)
        Cepstrum

    See Also
    --------
    pysptk.sptk.uels

    """

    return _sptk.fftcep(logsp, order, num_iter, acceleration_factor)


@apply_along_last_axis
@automatic_type_conversion
def lpc(windowed, order=25, min_det=1.0e-6, use_scipy=True):
    """Linear prediction analysis

    Parameters
    ----------
    windowed : array, shape (``frame_len``)
        A windowed frame

    order : int, optional
        Order of LPC. Default is 25.

    min_det : float, optional
        Mimimum value of the determinant of normal matrix.
        Default is 1.0e-6.

    use_scipy : bool
        Use scipy's solve_toeplitz implementation or not. Default is True.
        This would be more numerically stable than SPTK.

    Returns
    -------
    a : array, shape (``order + 1``)
        LPC

    Raises
    ------
    ValueError
        - if negative ``min_det`` is specified

    RuntimeError
        - if error happened in levdur


    See Also
    --------
    pysptk.sptk.lpc2par
    pysptk.sptk.par2lpc
    pysptk.sptk.lpc2c
    pysptk.sptk.lpc2lsp
    pysptk.sptk.ltcdf
    pysptk.sptk.lspdf

    """
    if use_scipy:
        r = _sptk.acorr(windowed, order)
        return levdur(r, use_scipy=True)
    else:
        return _sptk.lpc(windowed, order, min_det)


# MFCC #


@apply_along_last_axis
@automatic_type_conversion
def mfcc(
    x,
    order=14,
    fs=16000,
    alpha=0.97,
    eps=1.0,
    window_len=None,
    frame_len=None,
    num_filterbanks=20,
    cepslift=22,
    use_dft=False,
    use_hamming=False,
    czero=False,
    power=False,
):
    """MFCC

    Parameters
    ----------
    x : array
        A input signal

    order : int, optional
        Order of MFCC. Default is 14.

    fs : int, optional
        Sampling frequency. Default is 160000.

    alpha : float, optional
        Pre-emphasis coefficient. Default is 0.97.

    eps : float, optional
        Flooring value for calculating ``log(x)`` in filterbank analysis.
        Default is 1.0.

    window_len : int, optional
        Window lenght. Default is ``len(x)``.

    frame_len : int, optional
        Frame length. Default is ``len(x)``.

    num_filterbanks : int, optional
        Number of mel-filter banks. Default is 20.

    cepslift : int, optional
        Liftering coefficient. Default is 22.

    use_dft : bool, optional
        Use DFT (not FFT) or not. Default is False.

    use_hamming : bool, optional
        Use hamming window or not. Default is False.

    czero : bool, optional
        If True, ``mfcc`` returns 0-th coefficient as well. Default is False.

    power : bool, optional
        If True, ``mfcc`` returns power coefficient as well. Default is False.

    Returns
    -------
    cc : array
        MFCC vector, which is ordered as:

        mfcc[0], mfcc[1], mfcc[2], ... mfcc[order-1], c0, Power.

        Note that c0 and Power are optional.

        Shape of ``cc`` is:

            - ``order`` by default.
            - ``orde + 1`` if ``czero`` or ``power`` is set to True.
            - ``order + 2`` if both ``czero`` and ``power`` is set to True.

    Raises
    ------
    ValueError
        if ``num_filterbanks`` is less than or equal to ``order``

    See Also
    --------
    pysptk.sptk.gcep
    pysptk.sptk.mcep
    pysptk.sptk.mgcep

    """

    return _sptk.mfcc(
        x,
        order,
        fs,
        alpha,
        eps,
        window_len,
        frame_len,
        num_filterbanks,
        cepslift,
        use_dft,
        use_hamming,
        czero,
        power,
    )


# LPC, LSP and PARCOR conversions #


@apply_along_last_axis
@automatic_type_conversion
def lpc2c(lpc, order=None):
    """LPC to cepstrum

    Parameters
    ----------
    lpc : array
        LPC

    order : int, optional
        Order of cepstrum. Default is ``len(lpc) - 1``.

    Returns
    -------
    ceps : array, shape (``order + 1``)
        cepstrum

    See Also
    --------
    pysptk.sptk.lpc
    pysptk.sptk.lspdf

    """

    return _sptk.lpc2c(lpc, order)


@apply_along_last_axis
@automatic_type_conversion
def lpc2lsp(
    lpc,
    numsp=128,
    maxiter=4,
    eps=1.0e-6,
    has_gain=True,
    loggain=False,
    otype=0,
    fs=None,
):
    """LPC to LSP

    Parameters
    ----------
    lpc : array
        LPC

    numsp : int, optional
        Number of unit circle. Default is 128.

    maxiter : int, optional
        Maximum number of iteration. Default is 4.

    eps : float, optional
        End condition for iteration. Default is 1.0e-6.

    has_gain : bool, optional
        Whether input LPC has gain at the index 0 or not. Default is True.

    loggain : bool, optional
        whether the converted lsp should have loggain or not.
        Default is False.

    fs : int, optional
        Sampling frequency. Default is None and unused.

    otype : int, optional
        Output format LSP
            (0)  normalized frequency (0 ~ pi)
            (1)  normalized frequency (0 ~ 0.5)
            (2)  frequency (kHz)
            (3)  frequency (Hz)

        Default is 0.

    Returns
    -------
    lsp : array, shape (``order + 1``)
        LSP

    raises
    ------
    ValueError
        if ``fs`` is not specified when otype = 2 or 3.

    See Also
    --------
    pysptk.sptk.lpc
    pysptk.sptk.lspdf

    """

    return _sptk.lpc2lsp(lpc, numsp, maxiter, eps, has_gain, loggain, otype, fs)


@apply_along_last_axis
@automatic_type_conversion
def lpc2par(lpc):
    """LPC to PARCOR

    Parameters
    ----------
    lpc : array
        LPC

    Returns
    -------
    par : array, shape (same as ``lpc``)
        PARCOR

    See Also
    --------
    pysptk.sptk.lpc
    pysptk.sptk.par2lpc
    pysptk.sptk.ltcdf

    """

    return _sptk.lpc2par(lpc)


@apply_along_last_axis
@automatic_type_conversion
def par2lpc(par):
    """PARCOR to LPC

    Parameters
    ----------
    par : array
        PARCOR

    Returns
    -------
    lpc : array, shape (same as ``par``)
        LPC

    See Also
    --------
    pysptk.sptk.lpc
    pysptk.sptk.lpc2par

    """

    return _sptk.par2lpc(par)


@apply_along_last_axis
@automatic_type_conversion
def lsp2lpc(lsp, has_gain=True, loggain=False, fs=None, itype=0):
    """LSP to LPC


    Parameters
    ----------
    lpc : array
        LPC

    has_gain : bool, optional
        Whether input LPC has gain at the index 0 or not. Default is True.

    loggain : bool, optional
        If True, it's assumed that input LPC has loggain and convert it to linear
        gain. Default is False.

    fs : int, optional
        Sampling frequency. Default is None and unused.

    itype : int, optional
        Input LPC format
            (0)  normalized frequency (0 ~ pi)
            (1)  normalized frequency (0 ~ 0.5)
            (2)  frequency (kHz)
            (3)  frequency (Hz)

        Default is 0.

    Returns
    -------
    lsp : array, shape (``order + 1``) if has_gain else (``order``)
        LPC

    raises
    ------
    ValueError
        if ``fs`` is not specified when itype = 2 or 3.
        if ``loggain`` and not ``has_gain``.

    See Also
    --------
    pysptk.sptk.lpc2lsp
    """

    return _sptk.lsp2lpc(lsp, has_gain, loggain, fs, itype)


@apply_along_last_axis
@automatic_type_conversion
def lsp2sp(lsp, fftlen=256, has_gain=True, loggain=False, fs=None, itype=0):
    """LSP to spectrum

    Parameters
    ----------
    lsp : array
        LSP

    fftlen : int, optional
        FFT length

    has_gain : bool, optional
        Whether input LPC has gain at the index 0 or not. Default is True.

    loggain : bool, optional
        If True, it's assumed that input LPC has loggain and convert it to linear
        gain. Default is False.

    fs : int, optional
        Sampling frequency. Default is None and unused.

    itype : int, optional
        Input LPC format
            (0)  normalized frequency (0 ~ pi)
            (1)  normalized frequency (0 ~ 0.5)
            (2)  frequency (kHz)
            (3)  frequency (Hz)

        Default is 0.

    Returns
    -------
    sp : array, shape
        Spectrum. ln|H(z)|.

    See Also
    --------
    pysptk.sptk.lpc2par

    """

    return _sptk.lsp2sp(lsp, fftlen)


# Mel-generalized cepstrum conversions #


@apply_along_last_axis
@automatic_type_conversion
def mc2b(mc, alpha=0.35):
    """Mel-cepsrum to MLSA filter coefficients

    Parameters
    ----------
    mc : array, shape
        Mel-cepstrum.

    alpha : float, optional
        All-pass constant. Default is 0.35.

    Returns
    -------
    b : array, shape(same as ``mc``)
        MLSA filter coefficients

    See Also
    --------
    pysptk.sptk.mlsadf
    pysptk.sptk.mglsadf
    pysptk.sptk.b2mc
    pysptk.sptk.mcep
    pysptk.sptk.mgcep
    pysptk.sptk.amcep

    """

    return _sptk.mc2b(mc, alpha)


@apply_along_last_axis
@automatic_type_conversion
def b2mc(b, alpha=0.35):
    """MLSA filter coefficients to mel-cesptrum

    Parameters
    ----------
    b : array, shape
        MLSA filter coefficients

    alpha : float, optional
        All-pass constant. Default is 0.35.

    Returns
    -------
    mc : array, shape (same as ``b``)
        Mel-cepstrum.

    See Also
    --------
    pysptk.sptk.mc2b
    pysptk.sptk.mcep
    pysptk.sptk.mlsadf

    """

    return _sptk.b2mc(b, alpha)


@apply_along_last_axis
@automatic_type_conversion
def b2c(b, dst_order=None, alpha=0.35):
    return _sptk.b2c(b, dst_order, alpha)


@apply_along_last_axis
@automatic_type_conversion
def c2acr(c, order=None, fftlen=256):
    """Cepstrum to autocorrelation

    Parameters
    ----------
    c : array
        Cepstrum

    order : int, optional
        Order of cepstrum. Default is ``len(c) - 1``.

    fftlen : int, optional
        FFT length. Default is 256.

    Returns
    -------
    r : array, shape (``order + 1``)
        Autocorrelation

    Raises
    ------
    ValueError
        if non power of 2 ``fftlen`` is specified

    See Also
    --------
    pysptk.sptk.uels
    pysptk.sptk.c2ir
    pysptk.sptk.lpc2c
    pysptk.sptk.levdur

    """

    return _sptk.c2acr(c, order, fftlen)


@apply_along_last_axis
@automatic_type_conversion
def c2ir(c, length=256):
    """Cepstrum to impulse response

    Parameters
    ----------
    c : array
         Cepstrum

    length : int, optional
         Length of impulse response. Default is 256.

    Returns
    -------
    h : array, shape (``length``)
        impulse response

    See Also
    --------
    pysptk.sptk.c2acr

    """

    return _sptk.c2ir(c, length)


@apply_along_last_axis
@automatic_type_conversion
def ic2ir(h, order=25):
    """Impulse response to cepstrum

    Parameters
    ----------
    h : array
         Impulse response

    order : int, optional
         Order of cepstrum. Default is 25.

    Returns
    -------
    c : array, shape (``order + 1``)
        Cepstrum

    See Also
    --------
    pysptk.sptk.c2ir

    """

    return _sptk.ic2ir(h, order)


@apply_along_last_axis
@automatic_type_conversion
def c2ndps(c, fftlen=256):
    """Cepstrum to Negative Derivative of Phase Spectrum (NDPS)

    Parameters
    ----------
    c : array
         Cepstrum

    fftlen : int, optional
         FFT length. Default is 256.

    Returns
    -------
    ndps : array, shape (``fftlen // 2 + 1``)
        NDPS

    Raises
    ------
    ValueError
        if non power of 2 ``fftlen`` is specified

    See Also
    --------
    pysptk.sptk.mgcep
    pysptk.sptk.ndps2c

    """

    return _sptk.c2ndps(c, fftlen)


@apply_along_last_axis
@automatic_type_conversion
def ndps2c(ndps, order=25):
    """Cepstrum to Negative Derivative of Phase Spectrum (NDPS)

    Parameters
    ----------
    ndps : array, shape (``fftlen // 2 + 1``)
        NDPS

    order : int, optional
        Order of cepstrum. Default is 25.

    Returns
    -------
    c : array, shape (``order + 1``)
         Cepstrum

    Raises
    ------
    ValueError
        if non power of 2 ``fftlen`` is detected

    See Also
    --------
    pysptk.sptk.mgc2sp
    pysptk.sptk.c2ndps

    """

    return _sptk.ndps2c(ndps, order)


@apply_along_last_axis
@automatic_type_conversion
def gc2gc(src_ceps, src_gamma=0.0, dst_order=None, dst_gamma=0.0):
    """Generalized cepstrum transform

    Parameters
    ----------
    src_ceps : array
        Generalized cepstrum.

    src_gamma : float, optional
        Gamma of source cepstrum. Default is 0.0.

    dst_order : int, optional
        Order of destination cepstrum. Default is ``len(src_ceps) - 1``.

    dst_gamma : float, optional
        Gamma of destination cepstrum. Default is 0.0.

    Returns
    -------
    dst_ceps : array, shape (``dst_order + 1``)
         Converted generalized cepstrum

    Raises
    ------
    ValueError
        - if invalid ``src_gamma`` is specified
        - if invalid ``dst_gamma`` is specified

    See Also
    --------
    pysptk.sptk.gcep
    pysptk.sptk.mgcep
    pysptk.sptk.freqt
    pysptk.sptk.mgc2mgc
    pysptk.sptk.lpc2c

    """

    return _sptk.gc2gc(src_ceps, src_gamma, dst_order, dst_gamma)


@apply_along_last_axis
@automatic_type_conversion
def gnorm(ceps, gamma=0.0):
    """Gain normalization

    Parameters
    ----------
    ceps : array
        Generalized cepstrum.

    gamma : float, optional
        Gamma. Default is 0.0.

    Returns
    -------
    dst_ceps : array, shape(same as ``ceps``)
        Normalized generalized cepstrum

    Raises
    ------
    ValueError
        if invalid ``gamma`` is specified

    See Also
    --------
    pysptk.sptk.ignorm
    pysptk.sptk.gcep
    pysptk.sptk.mgcep
    pysptk.sptk.gc2gc
    pysptk.sptk.mgc2mgc
    pysptk.sptk.freqt

    """

    return _sptk.gnorm(ceps, gamma)


@apply_along_last_axis
@automatic_type_conversion
def ignorm(ceps, gamma=0.0):
    """Inverse gain normalization

    Parameters
    ----------
    c : array
        Normalized generalized cepstrum

    gamma : float, optional
        Gamma. Default is 0.0.

    Returns
    -------
    dst_ceps : array, shape (same as ``ceps``)
        Generalized cepstrum

    Raises
    ------
    ValueError
        if invalid ``gamma`` is specified

    See Also
    --------
    pysptk.sptk.gnorm
    pysptk.sptk.gcep
    pysptk.sptk.mgcep
    pysptk.sptk.gc2gc
    pysptk.sptk.mgc2mgc
    pysptk.sptk.freqt

    """

    return _sptk.ignorm(ceps, gamma)


@apply_along_last_axis
@automatic_type_conversion
def freqt(ceps, order=25, alpha=0.0):
    """Frequency transform

    Parameters
    ----------
    ceps : array
        Cepstrum.

    order : int, optional
        Desired order of transformed cepstrum. Default is 25.

    alpha : float, optional
        All-pass constant. Default is 0.0.

    Returns
    -------
    dst_ceps : array, shape(``order + 1``)
        frequency transofmed cepsttrum (typically mel-cepstrum)

    See Also
    --------
    pysptk.sptk.mgc2mgc

    """

    return _sptk.freqt(ceps, order, alpha)


@apply_along_last_axis
@automatic_type_conversion
def frqtr(src_ceps, order=25, alpha=0.0):
    return _sptk.frqtr(src_ceps, order, alpha)


@apply_along_last_axis
@automatic_type_conversion
def mgc2mgc(
    src_ceps, src_alpha=0.0, src_gamma=0.0, dst_order=None, dst_alpha=0.0, dst_gamma=0.0
):
    """Mel-generalized cepstrum transform

    Parameters
    ----------
    src_ceps : array
        Mel-generalized cepstrum.

    src_alpha : float, optional
        All-pass constant of source cesptrum. Default is 0.0.

    src_gamma : float, optional
        Gamma of source cepstrum. Default is 0.0.

    dst_order : int, optional
        Order of destination cepstrum. Default is ``len(src_ceps) - 1``.

    dst_alpha : float, optional
        All-pass constant of destination cesptrum. Default is 0.0.

    dst_gamma : float, optional
        Gamma of destination cepstrum. Default is 0.0.

    Returns
    -------
    dst_ceps : array, shape (``dst_order + 1``)
         Converted mel-generalized cepstrum

    Raises
    ------
    ValueError
        - if invalid ``src_gamma`` is specified
        - if invalid ``dst_gamma`` is specified

    See Also
    --------
    pysptk.sptk.uels
    pysptk.sptk.gcep
    pysptk.sptk.mcep
    pysptk.sptk.mgcep
    pysptk.sptk.gc2gc
    pysptk.sptk.freqt
    pysptk.sptk.lpc2c

    """

    return _sptk.mgc2mgc(
        src_ceps, src_alpha, src_gamma, dst_order, dst_alpha, dst_gamma
    )


@apply_along_last_axis
def mgc2sp(ceps, alpha=0.0, gamma=0.0, fftlen=256):
    """Mel-generalized cepstrum transform

    Parameters
    ----------
    ceps : array
        Mel-generalized cepstrum.

    alpha : float, optional
        All-pass constant. Default is 0.0.

    gamma : float, optional
        Gamma. Default is 0.0.

    fftlen : int, optional
        FFT length. Default is 256.

    Returns
    -------
    sp : array, shape (``fftlen // 2 + 1``)
         Complex spectrum log_e(H(ω))

    .. warning:: There's no equivalent option ``o`` in SPTK's mgc2sp.
        You have to normalize spectrum manually if needed.

    Raises
    ------
    ValueError
        - if invalid ``gamma`` is specified
        - if non power of 2 ``fftlen`` is specified

    See Also
    --------
    pysptk.sptk.mgc2mgc
    pysptk.sptk.gc2gc
    pysptk.sptk.freqt
    pysptk.sptk.gnorm
    pysptk.sptk.lpc2c

    """

    return _sptk.mgc2sp(ceps, alpha, gamma, fftlen)


@apply_along_last_axis
def mgclsp2sp(lsp, alpha=0.0, gamma=0.0, fftlen=256, gain=True):
    """MGC-LSP to spectrum

    Parameters
    ----------
    lsp : array
        MGC-LSP

    alpha : float, optional
        All-pass constant. Default is 0.0.

    gamma : float, optional
        Gamma. Default is 0.0.

    fftlen : int, optional
        FFT length. Default is 256.

    gain : bool, optional
        Whether the input MGC-LSP should have loggain or not.
        Default is True.

    Returns
    -------
    sp : array, shape (``fftlen // 2 + 1``)
         Complex spectrum

    Raises
    ------
    ValueError
        - if invalid ``gamma`` is specified
        - if non power of 2 ``fftlen`` is specified

    See Also
    --------
    pysptk.sptk.mgc2mgc

    """

    return _sptk.mgclsp2sp(lsp, alpha, gamma, fftlen, gain)


# F0 analysis #


@automatic_type_conversion
def swipe(x, fs, hopsize, min=60.0, max=240.0, threshold=0.3, otype="f0"):
    """SWIPE' - A Saw-tooth Waveform Inspired Pitch Estimation

    Parameters
    ----------
    x : array
        A whole audio signal

    fs : int
        Sampling frequency.

    hopsize : int
        Hop size.

    min : float, optional
        Minimum fundamental frequency. Default is 60.0

    max : float, optional
        Maximum fundamental frequency. Default is 240.0

    threshold : float, optional
        Voice/unvoiced threshold. Default is 0.3.

    otype : str or int, optional
        Output format
            (0) pitch
            (1) f0
            (2) log(f0)
        Default is f0.

    Returns
    -------
    f0  : array, shape(``np.ceil(float(len(x))/hopsize)``)
        Estimated f0 trajectory

    Raises
    ------
    ValueError
        if invalid otype is specified

    Examples
    --------

    >>> from scipy.io import wavfile
    >>> fs, x = wavfile.read(pysptk.util.example_audio_file())
    >>> hopsize = 80 # 5ms for 16kHz data
    >>> f0 = pysptk.swipe(x.astype(np.float64), fs, 80, otype="f0")

    >>> import matplotlib.pyplot as plt
    >>> plt.plot(f0, linewidth=2, label="F0 trajectory estimated by SWIPE'")
    >>> plt.xlim(0, len(f0))
    >>> plt.legend()
    >>> plt.tight_layout()

    See Also
    --------
    pysptk.sptk.rapt

    """

    return _sptk.swipe(x, fs, hopsize, min, max, threshold, otype)


@automatic_type_conversion_float32
def rapt(x, fs, hopsize, min=60, max=240, voice_bias=0.0, otype="f0"):
    """RAPT - a robust algorithm for pitch tracking

    Parameters
    ----------
    x : array, dtype=np.float32
        A whole audio signal

    fs : int
        Sampling frequency.

    hopsize : int
        Hop size.

    min : float, optional
        Minimum fundamental frequency. Default is 60.0

    max : float, optional
        Maximum fundamental frequency. Default is 240.0

    voice_bias : float, optional
        Voice/unvoiced threshold. Default is 0.0.

    otype : str or int, optional
        Output format
            (0) pitch
            (1) f0
            (2) log(f0)
        Default is f0.

    Notes
    -----
    It is assumed that input array ``x`` has np.float32 dtype, while swipe
    assumes np.float64 dtype.

    Returns
    -------
    f0  : array, shape(``np.ceil(float(len(x))/hopsize)``)
        Estimated f0 trajectory

    Raises
    ------
    ValueError
        - if invalid min/max frequency specified
        - if invalid frame period specified (not in [1/fs, 0.1])
        - if input range too small for analysis by get_f0

    RuntimeError
        - problem in init_dp_f0()

    Please see also the RAPT code in SPTK for more detailed exception conditions.

    Examples
    --------

    >>> from scipy.io import wavfile
    >>> fs, x = wavfile.read(pysptk.util.example_audio_file())
    >>> hopsize = 80 # 5ms for 16kHz data
    >>> f0 = pysptk.rapt(x.astype(np.float32), fs, 80, otype="f0")

    >>> import matplotlib.pyplot as plt
    >>> plt.plot(f0, linewidth=2, label="F0 trajectory estimated by RAPT")
    >>> plt.xlim(0, len(f0))
    >>> plt.legend()
    >>> plt.tight_layout()

    See Also
    --------
    pysptk.sptk.swipe

    """

    return _sptk.rapt(x, fs, hopsize, min, max, voice_bias, otype)


# Window functions #


def blackman(n, normalize=1):
    """Blackman window

    Parameters
    ----------
    n : int
         Window length

    normalize : int, optional
        Normalization flag
            (0) don't normalize
            (1) normalize by power
            (2) normalize by magnitude

        Default is 1.

    Returns
    -------
    w : array, shape (n,)
        blackman window

    """

    return _sptk.blackman(n, normalize)


def hamming(n, normalize=1):
    """Hamming window

    Parameters
    ----------
    n : int
         Window length

    normalize : int, optional
        Normalization flag
            (0) don't normalize
            (1) normalize by power
            (2) normalize by magnitude

        Default is 1.

    Returns
    -------
    w : array, shape (n,)
        hamming window

    """

    return _sptk.hamming(n, normalize)


def hanning(n, normalize=1):
    """Hanning window

    Parameters
    ----------
    n : int
         Window length

    normalize : int, optional
        Normalization flag
            (0) don't normalize
            (1) normalize by power
            (2) normalize by magnitude

        Default is 1.

    Returns
    -------
    w : array, shape (n,)
        hanning window

    """

    return _sptk.hanning(n, normalize)


def bartlett(n, normalize=1):
    """Bartlett window

    Parameters
    ----------
    n : int
         Window length

    normalize : int, optional
        Normalization flag
            (0) don't normalize
            (1) normalize by power
            (2) normalize by magnitude

        Default is 1.

    Returns
    -------
    w : array, shape (n,)
        bartlett window

    """

    return _sptk.bartlett(n, normalize)


def trapezoid(n, normalize=1):
    """Trapezoid window

    Parameters
    ----------
    n : int
         Window length

    normalize : int, optional
        Normalization flag
            (0) don't normalize
            (1) normalize by power
            (2) normalize by magnitude

        Default is 1.

    Returns
    -------
    w : array, shape (n,)
        trapezoid window

    """

    return _sptk.trapezoid(n, normalize)


def rectangular(n, normalize=1):
    """Rectangular window

    Parameters
    ----------
    n : int
         Window length

    normalize : int, optional
        Normalization flag
            (0) don't normalize
            (1) normalize by power
            (2) normalize by magnitude

        Default is 1.

    Returns
    -------
    w : array, shape (n,)
        rectangular window

    """

    return _sptk.rectangular(n, normalize)


# Waveform generation filters #


def zerodf_delay(order):
    """Delay for zerodf

    Parameters
    ----------
    order : int
        Order of zerodf filter coefficients

    Returns
    -------
    delay : array
        Delay

    """
    return np.zeros(_sptk.zerodf_delay_length(order))


def zerodf(x, b, delay):
    """All zero digital filter

    Parameters
    ----------
    x : float
        A input sample

    b : array
        FIR parameters

    delay : array
        Delay

    Returns
    -------
    y : float
        A filtered sample

    Raises
    ------
    ValueError
        if invalid delay length is supplied

    See Also
    --------
    pysptk.sptk.lpc
    pysptk.sptk.ltcdf
    pysptk.sptk.lmadf

    """

    return _sptk.zerodf(x, b, delay)


def zerodft(x, b, delay):
    """Transpose All zero digital filter

    Parameters
    ----------
    x : float
        A input sample

    b : array
        FIR parameters

    delay : array
        Delay

    Returns
    -------
    y : float
        A filtered sample

    Raises
    ------
    ValueError
        if invalid delay length is supplied

    See Also
    --------
    pysptk.sptk.zerodf

    """

    return _sptk.zerodft(x, b, delay)


def poledf_delay(order):
    """Delay for poledf

    Parameters
    ----------
    order : int
        Order of poledf filter coefficients

    Returns
    -------
    delay : array
        Delay

    """
    return np.zeros(_sptk.poledf_delay_length(order))


def poledf(x, a, delay):
    """All-pole digital filter

    Parameters
    ----------
    x : float
        A input sample

    a : array
        AR coefficients

    delay : array
        Delay

    Returns
    -------
    y : float
        A filtered sample

    Raises
    ------
    ValueError
        if invalid delay length is supplied

    See Also
    --------
    pysptk.sptk.lpc
    pysptk.sptk.ltcdf
    pysptk.sptk.lmadf

    """

    return _sptk.poledf(x, a, delay)


def poledft(x, a, delay):
    """Transpose All-pole digital filter

    Parameters
    ----------
    x : float
        A input sample

    a : array
        AR coefficients

    delay : array
        Delay

    Returns
    -------
    y : float
        A filtered sample

    Raises
    ------
    ValueError
        if invalid delay length is supplied

    See Also
    --------
    pysptk.sptk.poledf

    """

    return _sptk.poledft(x, a, delay)


def lmadf_delay(order, pd):
    """Delay for lmadf

    Parameters
    ----------
    order : int
        Order of lmadf filter coefficients

    pd : int
        Order of pade approximation.

    Returns
    -------
    delay : array
        Delay

    """

    return np.zeros(_sptk.lmadf_delay_length(order, pd))


def lmadf(x, b, pd, delay):
    """LMA digital filter

    Parameters
    ----------
    x : float
        A input sample

    c : array
        Cepstrum

    pd : int
        Order of pade approximation

    delay : array
        Delay

    Returns
    -------
    y : float
        A filtered sample

    Raises
    ------
    ValueError
        - if invalid order of pade approximation is specified
        - if invalid delay length is supplied

    See Also
    --------
    pysptk.sptk.uels
    pysptk.sptk.acep
    pysptk.sptk.poledf
    pysptk.sptk.ltcdf
    pysptk.sptk.glsadf
    pysptk.sptk.mlsadf
    pysptk.sptk.mglsadf

    """

    return _sptk.lmadf(x, b, pd, delay)


def lspdf_delay(order):
    """Delay for lspdf

    Parameters
    ----------
    order : int
        Order of lspdf filter coefficients

    Returns
    -------
    delay : array
        Delay

    """

    return np.zeros(_sptk.lspdf_delay_length(order))


def lspdf(x, f, delay):
    """LSP synthesis digital filter

    Parameters
    ----------
    x : float
        A input sample

    f : array
        LSP coefficients

    delay : array
        Delay

    Returns
    -------
    y : float
        A filtered sample

    Raises
    ------
    ValueError
        if invalid delay length is supplied

    See Also
    --------
    pysptk.sptk.lpc2lsp

    """

    return _sptk.lspdf(x, f, delay)


def ltcdf_delay(order):
    """Delay for ltcdf

    Parameters
    ----------
    order : int
        Order of ltcdf filter coefficients

    Returns
    -------
    delay : array
        Delay

    """

    return np.zeros(_sptk.ltcdf_delay_length(order))


def ltcdf(x, k, delay):
    """All-pole lattice digital filter

    Parameters
    ----------
    x : float
        A input sample

    k : array
        PARCOR coefficients.

    delay : array
        Delay

    Returns
    -------
    y : float
        A filtered sample

    Raises
    ------
    ValueError
        if invalid delay length is supplied

    See Also
    --------
    pysptk.sptk.lpc
    pysptk.sptk.lpc2par
    pysptk.sptk.lpc2lsp
    pysptk.sptk.poledf
    pysptk.sptk.lspdf

    """

    return _sptk.ltcdf(x, k, delay)


def glsadf_delay(order, stage):
    """Delay for glsadf

    Parameters
    ----------
    order : int
        Order of glsadf filter coefficients

    stage : int
        -1 / gamma

    Returns
    -------
    delay : array
        Delay

    """

    return np.zeros(_sptk.glsadf_delay_length(order, stage))


def glsadf(x, c, stage, delay):
    """GLSA digital filter

    Parameters
    ----------
    x : float
        A input sample

    c : array
        Geneeraized cepstrum

    stage : int
        -1 / gamma

    delay : array
        Delay

    Returns
    -------
    y : float
        A filtered sample

    Raises
    ------
    ValueError
        - if invalid number of stage is specified
        - if invalid delay length is supplied

    See Also
    --------
    pysptk.sptk.ltcdf
    pysptk.sptk.lmadf
    pysptk.sptk.lspdf
    pysptk.sptk.mlsadf
    pysptk.sptk.mglsadf

    """

    return _sptk.glsadf(x, c, stage, delay)


def glsadft(x, c, stage, delay):
    """Transpose GLSA digital filter

    Parameters
    ----------
    x : float
        A input sample

    c : array
        Geneeraized cepstrum

    stage : int
        -1 / gamma

    delay : array
        Delay

    Returns
    -------
    y : float
        A filtered sample

    Raises
    ------
    ValueError
        - if invalid number of stage is specified
        - if invalid delay length is supplied

    See Also
    --------
    pysptk.sptk.glsadf
    """

    return _sptk.glsadft(x, c, stage, delay)


def mlsadf_delay(order, pd):
    """Delay for mlsadf

    Parameters
    ----------
    order : int
        Order of mlsadf filter coefficients

    pd : int
        Order of pade approximation.

    Returns
    -------
    delay : array
        Delay

    """

    return np.zeros(_sptk.mlsadf_delay_length(order, pd))


def mlsadf(x, b, alpha, pd, delay):
    """MLSA digital filter

    Parameters
    ----------
    x : float
        A input sample

    b : array
        MLSA filter coefficients

    alpha : float
        All-pass constant

    pd : int
        Order of pade approximation

    delay : array
        Delay

    Returns
    -------
    y : float
        A filtered sample

    Raises
    ------
    ValueError
        - if invalid order of pade approximation is specified
        - if invalid delay length is supplied

    See Also
    --------
    pysptk.sptk.mcep
    pysptk.sptk.amcep
    pysptk.sptk.poledf
    pysptk.sptk.ltcdf
    pysptk.sptk.lmadf
    pysptk.sptk.lspdf
    pysptk.sptk.glsadf
    pysptk.sptk.mglsadf

    """

    return _sptk.mlsadf(x, b, alpha, pd, delay)


def mlsadft(x, b, alpha, pd, delay):
    """Transpose MLSA digital filter

    Parameters
    ----------
    x : float
        A input sample

    b : array
        MLSA filter coefficients

    alpha : float
        All-pass constant

    pd : int
        Order of pade approximation

    delay : array
        Delay

    Returns
    -------
    y : float
        A filtered sample

    Raises
    ------
    ValueError
        - if invalid order of pade approximation is specified
        - if invalid delay length is supplied

    See Also
    --------
    pysptk.sptk.mlsadf

    """

    return _sptk.mlsadft(x, b, alpha, pd, delay)


def mglsadf_delay(order, stage):
    """Delay for mglsadf

    Parameters
    ----------
    order : int
        Order of mglsadf filter coefficients

    stage : int
        -1 / gamma

    Returns
    -------
    delay : array
        Delay

    """

    return np.zeros(_sptk.mglsadf_delay_length(order, stage))


def mglsadf(x, b, alpha, stage, delay):
    """MGLSA digital filter

    Parameters
    ----------
    x : float
        A input sample

    b : array
        MGLSA filter coefficients

    alpha : float
        All-pass constant

    stage : int
        -1 / gamma

    delay : array
        Delay

    Returns
    -------
    y : float
        A filtered sample

    Raises
    ------
    ValueError
        - if invalid number of stage is specified
        - if invalid delay length is supplied

    See Also
    --------
    pysptk.sptk.mgcep
    pysptk.sptk.poledf
    pysptk.sptk.ltcdf
    pysptk.sptk.lmadf
    pysptk.sptk.lspdf
    pysptk.sptk.mlsadf
    pysptk.sptk.glsadf

    """

    return _sptk.mglsadf(x, b, alpha, stage, delay)


def mglsadft(x, b, alpha, stage, delay):
    """Transpose MGLSA digital filter

    Parameters
    ----------
    x : float
        A input sample

    b : array
        MGLSA filter coefficients

    alpha : float
        All-pass constant

    stage : int
        -1 / gamma

    delay : array
        Delay

    Returns
    -------
    y : float
        A filtered sample

    Raises
    ------
    ValueError
        - if invalid number of stage is specified
        - if invalid delay length is supplied

    See Also
    --------
    pysptk.sptk.mglsadf
    """

    return _sptk.mglsadft(x, b, alpha, stage, delay)


# Excitation #


def excite(pitch, hopsize=100, interp_period=1, gaussian=False, seed=1):
    """Excitation generation

    Parameters
    ----------
    pitch : array
        Pitch sequence.

        .. note::

            ``excite`` assumes that input is a **pitch** sequence, not **f0**
            sequence. Pitch sequence can be obtained by speficying
            ```otype="pitch"`` to F0 estimation methods.

    hopsize : int
        Hop size (frame period in sample). Default is 100.

    interp_period : int
        Interpolation period. Default is 1.

    gaussian : bool
        If True, generate gausssian noise for unvoiced frames, otherwise
        generate M-sequence. Default is False.

    seed : int
        Seed for nrand for Gaussian noise. Default is 1.

    Returns
    -------
    excitation : array
        Excitation signal

    See also
    --------

    pysptk.sptk.poledf
    pysptk.sptk.swipe
    pysptk.sptk.rapt


    """
    return _sptk.excite(pitch, hopsize, interp_period, gaussian, seed)


# Utils #


def phidf(x, order, alpha, delay):
    _sptk.phidf(x, order, alpha, delay)


def lspcheck(lsp):
    return _sptk.lspcheck(lsp)


@apply_along_last_axis
@automatic_type_conversion
def levdur(r, eps=0.0, use_scipy=True):
    """Solve an Autocorrelation Normal Equation Using Levinson-Durbin Method

    Parameters
    ----------
    r : array
        Autocorrelation sequence

    eps : float
        Singular check (eps(if -1., 0.0 is assumed))

    use_scipy : bool
        Use scipy's solve_toeplitz implementation or not. Default is True.
        This would be more numerically stable than SPTK.

    Returns
    -------
    a : array
        LPC

    Raises
    ------
    RuntimeError
        - if abnormally completed or unstable LPC

    See Also
    --------
    pysptk.sptk.c2acr
    """
    if use_scipy:
        from scipy.linalg import solve_toeplitz

        a = np.empty_like(r)
        # Rx = r where R is a toeplitz matrix
        a[1:] = -solve_toeplitz(r[:-1], r[1:])
        # http://www.seas.ucla.edu/~ingrid/ee213a/speech/vlad_present.pdf
        G = np.sqrt(r[0] + (a[1:] * r[1:]).sum())
        a[0] = G
        return a
    else:
        return _sptk.levdur(r, eps)



import numpy as np
import pysptk
import pytest


def windowed_dummy_data(N):
    np.random.seed(98765)
    return pysptk.hanning(N) * np.random.randn(N)


def windowed_dummy_frames(T, N, dtype=np.float64):
    np.random.seed(98765)
    frames = pysptk.hanning(N) * np.random.randn(T, N)
    return frames.astype(np.float64)


@pytest.mark.parametrize("order", [15, 20, 25])
@pytest.mark.parametrize("alpha", [0.0, 0.35, 0.41])
def test_mcep(order, alpha):
    x = windowed_dummy_data(1024)

    def __test(order, alpha):
        mc = pysptk.mcep(x, order, alpha)
        assert np.all(np.isfinite(mc))

    __test(order, alpha)


def test_mcep_broadcast():
    # Test for broadcasting
    def __test_broadcast(dtype):
        frames = windowed_dummy_frames(100, 512, dtype=dtype)
        mc = pysptk.mcep(frames, 20, 0.41)
        assert np.all(np.isfinite(mc))
        assert frames.shape[0] == mc.shape[0]

    for dtype in [np.float16, np.float32, np.float64]:
        __test_broadcast(dtype)


def test_mcep_invalid_args():
    x = windowed_dummy_data(1024)

    def __test_itype(itype=0):
        pysptk.mcep(x, itype=itype)

    with pytest.raises(ValueError):
        __test_itype(-1)
    with pytest.raises(ValueError):
        __test_itype(5)

    def __test_eps(etype=0, eps=0.0):
        pysptk.mcep(x, etype=etype, eps=eps)

    with pytest.raises(ValueError):
        __test_eps(0, -1.0)
    with pytest.raises(ValueError):
        __test_eps(-1)
    with pytest.raises(ValueError):
        __test_eps(-3)
    with pytest.raises(ValueError):
        __test_eps(1, -1.0)
    with pytest.raises(ValueError):
        __test_eps(2, 0.0)
    with pytest.raises(ValueError):
        __test_eps(2, 1.0)

    def __test_min_det(min_det):
        pysptk.mcep(x, min_det=min_det)

    with pytest.raises(ValueError):
        __test_min_det(-1.0)


def test_mcep_failure():
    with pytest.raises(RuntimeError):
        pysptk.mcep(np.ones(256), 40, 0.41)


@pytest.mark.parametrize("order", [15, 20, 25])
@pytest.mark.parametrize("gamma", [0.0, -0.25, -0.5])
def test_gcep(order, gamma):
    x = windowed_dummy_data(1024)
    gc = pysptk.gcep(x, order, gamma)
    assert np.all(np.isfinite(gc))


def test_gcep_invalid_args():
    x = windowed_dummy_data(1024)

    def __test_gamma(gamma):
        pysptk.gcep(x, gamma=gamma)

    with pytest.raises(ValueError):
        __test_gamma(0.1)
    with pytest.raises(ValueError):
        __test_gamma(-2.1)

    def __test_itype(itype=0):
        pysptk.gcep(x, itype=itype)

    with pytest.raises(ValueError):
        __test_itype(-1)
    with pytest.raises(ValueError):
        __test_itype(5)

    def __test_eps(etype=0, eps=0.0):
        pysptk.gcep(x, etype=etype, eps=eps)

    with pytest.raises(ValueError):
        __test_eps(0, -1.0)
    with pytest.raises(ValueError):
        __test_eps(-1)
    with pytest.raises(ValueError):
        __test_eps(-3)
    with pytest.raises(ValueError):
        __test_eps(1, -1.0)
    with pytest.raises(ValueError):
        __test_eps(2, -1.0)

    def __test_min_det(min_det):
        pysptk.gcep(x, min_det=min_det)

    with pytest.raises(ValueError):
        __test_min_det(-1.0)


def test_gcep_failure():
    with pytest.raises(RuntimeError):
        pysptk.gcep(np.ones(256), 40, 0.0)


@pytest.mark.parametrize("order", [15, 20, 25])
@pytest.mark.parametrize("alpha", [0.0, 0.35, 0.41])
@pytest.mark.parametrize("gamma", [-0.5, -0.25, 0.0])
def test_mgcep(order, alpha, gamma):
    x = windowed_dummy_data(1024)

    mgc = pysptk.mgcep(x, order, alpha, gamma)
    assert np.all(np.isfinite(mgc))


def test_mgcep_invalid_args():
    x = windowed_dummy_data(1024)

    def __test_gamma(gamma):
        pysptk.mgcep(x, gamma=gamma)

    with pytest.raises(ValueError):
        __test_gamma(0.1)
    with pytest.raises(ValueError):
        __test_gamma(-2.1)

    def __test_itype(itype=0):
        pysptk.mgcep(x, itype=itype)

    with pytest.raises(ValueError):
        __test_itype(-1)
    with pytest.raises(ValueError):
        __test_itype(5)

    def __test_eps(etype=0, eps=0.0):
        pysptk.mgcep(x, etype=etype, eps=eps)

    with pytest.raises(ValueError):
        __test_eps(0, -1.0)
    with pytest.raises(ValueError):
        __test_eps(-1)
    with pytest.raises(ValueError):
        __test_eps(-3)
    with pytest.raises(ValueError):
        __test_eps(1, -1.0)
    with pytest.raises(ValueError):
        __test_eps(2, -1.0)

    def __test_min_det(min_det):
        pysptk.mgcep(x, min_det=min_det)

    with pytest.raises(ValueError):
        __test_min_det(-1.0)

    def __test_otype(otype=0):
        pysptk.mgcep(x, otype=otype)

    with pytest.raises(ValueError):
        __test_otype(-1)
    with pytest.raises(ValueError):
        __test_otype(6)


def test_mgcep_failure():
    with pytest.raises(RuntimeError):
        pysptk.mgcep(np.ones(256))


@pytest.mark.parametrize("order", [15, 20, 25])
def test_uels(order):
    x = windowed_dummy_data(1024)

    c = pysptk.uels(x, order)
    assert np.all(np.isfinite(c))


def test_uels_invalid_args():
    x = windowed_dummy_data(1024)

    def __test_itype(itype=0):
        pysptk.uels(x, itype=itype)

    with pytest.raises(ValueError):
        __test_itype(-1)
    with pytest.raises(ValueError):
        __test_itype(5)

    def __test_eps(etype=0, eps=0.0):
        pysptk.uels(x, etype=etype, eps=eps)

    with pytest.raises(ValueError):
        __test_eps(0, -1.0)
    with pytest.raises(ValueError):
        __test_eps(-1)
    with pytest.raises(ValueError):
        __test_eps(-3)
    with pytest.raises(ValueError):
        __test_eps(1, -1.0)
    with pytest.raises(ValueError):
        __test_eps(2, -1.0)


def test_uels_failure():
    with pytest.raises(RuntimeError):
        pysptk.uels(np.ones(256), 40)


@pytest.mark.parametrize("order", [15, 20, 25])
def test_fftcep(order):
    x = windowed_dummy_data(1024)
    logsp = np.log(np.abs(np.fft.rfft(x)) + 1.0e-6)

    c = pysptk.fftcep(logsp, order)
    assert np.all(np.isfinite(c))


@pytest.mark.parametrize("order", [15, 20, 25])
def test_lpc(order):
    x = windowed_dummy_data(1024)

    a1 = pysptk.lpc(x, order, use_scipy=False)
    a2 = pysptk.lpc(x, order, use_scipy=True)
    a3 = pysptk.levdur(pysptk.acorr(x, order), use_scipy=False)
    a4 = pysptk.levdur(pysptk.acorr(x, order), use_scipy=True)

    assert np.all(np.isfinite(a1))
    assert np.allclose(a1, a2)
    assert np.allclose(a1, a3)
    assert np.allclose(a1, a4)


def test_lpc_invalid_args():
    x = windowed_dummy_data(1024)

    def __test_min_det(min_det):
        pysptk.lpc(x, min_det=min_det, use_scipy=False)

    with pytest.raises(ValueError):
        __test_min_det(-1.0)


def test_lpc_failure():
    with pytest.raises(RuntimeError):
        pysptk.lpc(np.zeros(256), 40, use_scipy=False)



import numpy as np
import pysptk
import pytest


def test_mfcc_options():
    np.random.seed(98765)
    dummy_input = np.random.rand(512)

    # with c0
    cc = pysptk.mfcc(dummy_input, 12, czero=True)
    assert len(cc) == 13

    # wth c0 + power
    cc = pysptk.mfcc(dummy_input, 12, czero=True, power=True)
    assert len(cc) == 14

    # with power
    cc = pysptk.mfcc(dummy_input, 12, power=True)
    assert len(cc) == 13


@pytest.mark.parametrize("n", [21, 23, 25])
def test_mfcc_num_filterbanks(n):
    def __test(n):
        np.random.seed(98765)
        dummy_input = np.random.rand(512)
        cc = pysptk.mfcc(dummy_input, 20, num_filterbanks=n)
        assert np.all(np.isfinite(cc))

    __test(n)


def test_mgcc_num_filterbanks_corner_case():
    def __test(n):
        np.random.seed(98765)
        dummy_input = np.random.rand(512)
        pysptk.mfcc(dummy_input, 20, num_filterbanks=n)

    for n in [19, 20]:
        with pytest.raises(ValueError):
            __test(n)


@pytest.mark.parametrize("order", [12, 14, 16, 18])
@pytest.mark.parametrize("length", [256, 512, 1024, 2048, 4096])
def test_mfcc(order, length):
    def __test(length, order):
        np.random.seed(98765)
        dummy_input = np.random.rand(length)
        cc = pysptk.mfcc(dummy_input, order, czero=True, power=True)
        assert np.all(np.isfinite(cc))

    __test(length, order)



# coding: utf-8

"""
Other conversions
-----------------

Not exist in SPTK itself, but can be used with the core API.
Functions in the ``pysptk.conversion`` module can also be directly accesible by ``pysptk.*``.

.. autosummary::
    :toctree: generated/

    mgc2b
    sp2mc
    mc2sp
    mc2e
"""

from __future__ import absolute_import, division, print_function

import numpy as np
from pysptk.sptk import c2ir, freqt, gnorm, mc2b
from pysptk.util import apply_along_last_axis, automatic_type_conversion


@apply_along_last_axis
@automatic_type_conversion
def mgc2b(mgc, alpha=0.35, gamma=0.0):
    """Mel-generalized cepstrum to MGLSA filter coefficients

    Parameters
    ----------
    mgc : array, shape
        Mel-generalized cepstrum

    alpha : float
        All-pass constant. Default is 0.35.

    gamma : float
        Parameter of generalized log function. Default is 0.0.

    Returns
    -------
    b : array, shape(same as ``mgc``)
        MGLSA filter coefficients

    See Also
    --------
    pysptk.sptk.mlsadf
    pysptk.sptk.mglsadf
    pysptk.sptk.mc2b
    pysptk.sptk.b2mc
    pysptk.sptk.mcep
    pysptk.sptk.mgcep

    """

    b = mc2b(mgc, alpha)
    if gamma == 0:
        return b

    b = gnorm(b, gamma)

    b[0] = np.log(b[0])
    b[1:] *= gamma

    return b


@apply_along_last_axis
@automatic_type_conversion
def sp2mc(powerspec, order, alpha):
    """Convert spectrum envelope to mel-cepstrum

    This is a simplified implementation of ``mcep`` for input type
    is 4.

    Parameters
    ----------
    powerspec : array
        Power spectrum

    order : int
        Order of mel-cepstrum

    alpha : float
        All-pass constant.

    Returns
    -------
    mc : array, shape(``order+1``)
        mel-cepstrum

    See Also
    --------
    pysptk.sptk.mcep
    pysptk.conversion.mc2sp

    """

    # |X(ω)|² -> log(|X(ω)²|)
    logperiodogram = np.log(powerspec)

    # transform log-periodogram to real cepstrum
    # log(|X(ω)|²) -> c(m)
    c = np.fft.irfft(logperiodogram)
    c[0] /= 2.0

    # c(m) -> cₐ(m)
    return freqt(c, order, alpha)


@apply_along_last_axis
@automatic_type_conversion
def mc2sp(mc, alpha, fftlen):
    """Convert mel-cepstrum back to power spectrum

    Parameters
    ----------
    mc : array
        Mel-spectrum

    alpha : float
        All-pass constant.

    fftlen : int
        FFT length

    Returns
    -------
    powerspec : array, shape(``fftlen//2 +1``)
        Power spectrum

    See Also
    --------
    pysptk.sptk.mcep
    pysptk.conversion.sp2mc

    """
    # back to cepstrum from mel-cesptrum
    # cₐ(m) -> c(m)
    c = freqt(mc, int(fftlen // 2), -alpha)
    c[0] *= 2.0

    symc = np.zeros(fftlen)
    symc[0] = c[0]
    for i in range(1, len(c)):
        symc[i] = c[i]
        symc[-i] = c[i]

    # back to power spectrum
    # c(m) -> log(|X(ω)|²) -> |X(ω)|²
    return np.exp(np.fft.rfft(symc).real)


@apply_along_last_axis
@automatic_type_conversion
def mc2e(mc, alpha=0.35, irlen=256):
    """Compute energy from mel-cepstrum

    Inspired from hts_engine

    Parameters
    ----------
    mc : array
        Mel-spectrum

    alpha : float
        All-pass constant.

    irlen : int
        IIR filter length

    Returns
    -------
    energy : floating point, scalar
        frame energy
    """
    # back to linear frequency domain
    c = freqt(mc, irlen - 1, -alpha)

    # compute impule response from cepsturm
    ir = c2ir(c, irlen)

    return np.sum(np.abs(ir * ir))



import numpy as np
import pysptk


def test_agexp():
    assert pysptk.agexp(1, 1, 1) == 5.0
    assert pysptk.agexp(1, 2, 3) == 18.0


def test_gexp():
    assert pysptk.gexp(1, 1) == 2.0
    assert pysptk.gexp(2, 4) == 3.0


def test_glog():
    assert pysptk.glog(1, 2) == 1.0
    assert pysptk.glog(2, 3) == 4.0


def test_mseq():
    for _ in range(0, 100):
        assert np.isfinite(pysptk.mseq())



from warnings import warn

import numpy as np
import pysptk
import pytest


def windowed_dummy_data(N):
    np.random.seed(98765)
    return pysptk.hanning(N) * np.random.randn(N)


def __test_conversion_base(f, src_order, dst_order, *args, **kwargs):
    np.random.seed(98765)
    src = np.random.rand(src_order + 1)
    dst = f(src, dst_order, *args, **kwargs)

    assert np.all(np.isfinite(dst))


def __test_transform_base(f, order, *args, **kwargs):
    np.random.seed(98765)
    src = np.random.rand(order + 1)
    dst = f(src, *args, **kwargs)

    assert np.all(np.isfinite(dst))


@pytest.mark.parametrize("src_order", [15, 20, 25, 30])
@pytest.mark.parametrize("dst_order", [15, 20, 25, 30])
def test_lpc2c(src_order, dst_order):
    __test_conversion_base(pysptk.lpc2c, src_order, dst_order)


def test_lpc2lsp():
    # for order in [15, 20, 25, 30]:
    #    yield __test_transform_base, pysptk.lpc2lsp, order

    def __test_invalid_otype(dummy_lpc, otype):
        pysptk.lpc2lsp(dummy_lpc, otype=otype)

    np.random.seed(98765)
    dummy_lpc = pysptk.lpc(np.random.rand(512), 21)

    # invalid otype
    with pytest.raises(ValueError):
        __test_invalid_otype(dummy_lpc, 2)
    with pytest.raises(ValueError):
        __test_invalid_otype(dummy_lpc, 3)

    lsp1 = pysptk.lpc2lsp(dummy_lpc, otype=2, fs=16000)
    lsp2 = pysptk.lpc2lsp(dummy_lpc, otype=3, fs=16)
    assert np.allclose(lsp1, lsp2)

    # loggain
    lsp3 = pysptk.lpc2lsp(dummy_lpc, otype=3, fs=16, loggain=True)
    assert lsp3[0] == np.log(lsp2[0])


@pytest.mark.parametrize("order", [15, 20, 25, 30])
def test_lpc2par(order):
    __test_transform_base(pysptk.lpc2par, order)


@pytest.mark.parametrize("order", [15, 20, 25, 30])
def test_par2lpc(order):
    __test_transform_base(pysptk.par2lpc, order)


@pytest.mark.parametrize("order", [15, 20, 25, 30])
@pytest.mark.parametrize("fftlen", [256, 512, 1024])
def test_lsp2sp(order, fftlen):
    __test_transform_base(pysptk.lsp2sp, order, fftlen)


def test_lsp2sp_fftlen():
    def __test_fftlen(fftlen):
        pysptk.lsp2sp(np.ones(20), fftlen)

    with pytest.raises(ValueError):
        __test_fftlen(257)
    with pytest.raises(ValueError):
        __test_fftlen(513)


@pytest.mark.parametrize("order", [15, 20, 25, 30])
@pytest.mark.parametrize("alpha", [0.35, 0.41, 0.5])
def test_mc2b(order, alpha):
    __test_transform_base(pysptk.mc2b, order, alpha)


@pytest.mark.parametrize("order", [15, 20, 25, 30])
@pytest.mark.parametrize("alpha", [0.35, 0.41, 0.5])
def test_b2mc(order, alpha):
    __test_transform_base(pysptk.b2mc, order, alpha)


@pytest.mark.parametrize("src_order", [15, 20, 25, 30])
@pytest.mark.parametrize("dst_order", [15, 20, 25, 30])
@pytest.mark.parametrize("alpha", [0.35, 0.41, 0.5])
def test_b2c(src_order, dst_order, alpha):
    __test_conversion_base(pysptk.b2c, src_order, dst_order, alpha)


@pytest.mark.parametrize("src_order", [15, 20, 25, 30])
@pytest.mark.parametrize("dst_order", [15, 20, 25, 30])
@pytest.mark.parametrize("fftlen", [256, 512, 1024])
def test_c2acr(src_order, dst_order, fftlen):
    __test_transform_base(pysptk.b2c, src_order, dst_order, fftlen)


def test_c2acr_fftlen():
    def __test_fftlen(fftlen):
        pysptk.c2acr(np.ones(20), 19, fftlen)

    with pytest.raises(ValueError):
        __test_fftlen(257)
    with pytest.raises(ValueError):
        __test_fftlen(513)
    with pytest.raises(ValueError):
        __test_fftlen(16)


@pytest.mark.parametrize("order", [15, 20, 25, 30])
@pytest.mark.parametrize("length", [256, 512, 1024])
def test_c2ir(order, length):
    __test_conversion_base(pysptk.c2ir, order, length)


@pytest.mark.parametrize("order", [15, 20, 25, 30])
@pytest.mark.parametrize("length", [256, 512, 1024])
def test_ic2ir(order, length):
    __test_conversion_base(pysptk.ic2ir, length, order)


@pytest.mark.parametrize("order", [15, 20, 25, 30])
@pytest.mark.parametrize("length", [256, 512, 1024])
def test_ic2ir_invertibility(order, length):
    np.random.seed(98765)
    dummy_ceps = np.random.rand(order + 1)
    ir = pysptk.c2ir(dummy_ceps, length)
    c = pysptk.ic2ir(ir, order)
    assert np.allclose(c, dummy_ceps)


@pytest.mark.parametrize("order", [15, 20, 25, 30])
@pytest.mark.parametrize("fftlen", [256, 512, 1024])
def test_c2ndps(order, fftlen):
    __test_conversion_base(pysptk.c2ndps, order, fftlen)


def test_c2ndps_fftlen():
    # invalid fftlen
    with pytest.raises(ValueError):
        __test_conversion_base(pysptk.c2ndps, 20, 255)
    with pytest.raises(ValueError):
        __test_conversion_base(pysptk.c2ndps, 20, 257)


@pytest.mark.parametrize("order", [15, 20, 25, 30])
@pytest.mark.parametrize("fftlen", [256, 512, 1024])
def test_ndps2c(order, fftlen):
    __test_conversion_base(pysptk.ndps2c, (fftlen >> 1), order)


def test_ndps2c_fftlen():
    def __test(length):
        pysptk.ndps2c(np.ones(length), 20)

    # invalid fftlen
    with pytest.raises(ValueError):
        __test(255 >> 1)
    with pytest.raises(ValueError):
        __test(257 >> 1)


@pytest.mark.parametrize("src_order", [15, 20, 25, 30])
@pytest.mark.parametrize("src_gamma", [-1.0, -0.5, 0.0])
@pytest.mark.parametrize("dst_order", [15, 20, 25, 30])
@pytest.mark.parametrize("dst_gamma", [-1.0, -0.5, 0.0])
def test_gc2gc(src_order, src_gamma, dst_order, dst_gamma):
    np.random.seed(98765)
    src = np.random.rand(src_order + 1)
    dst = pysptk.gc2gc(src, src_gamma, dst_order, dst_gamma)
    assert np.all(np.isfinite(dst))


def test_gc2gc_gamma():
    def __test(src_order, src_gamma, dst_order, dst_gamma):
        np.random.seed(98765)
        src = np.random.rand(src_order + 1)
        dst = pysptk.gc2gc(src, src_gamma, dst_order, dst_gamma)
        assert np.all(np.isfinite(dst))

    # invalid gamma
    with pytest.raises(ValueError):
        __test(20, 0.0, 20, 0.1)
    with pytest.raises(ValueError):
        __test(20, 0.1, 20, 0.0)


@pytest.mark.parametrize("order", [15, 20, 25, 30])
@pytest.mark.parametrize("gamma", [-1.0, -0.5, 0.0])
def test_gnorm(order, gamma):
    __test_transform_base(pysptk.gnorm, order, gamma)


def test_gnorm_gamma():
    # invalid gamma
    with pytest.raises(ValueError):
        __test_transform_base(pysptk.gnorm, 20, 0.1)


@pytest.mark.parametrize("order", [15, 20, 25, 30])
@pytest.mark.parametrize("gamma", [-1.0, -0.5, 0.0])
def test_ignorm(order, gamma):
    __test_transform_base(pysptk.ignorm, order, gamma)


def test_ignorm_gamma():
    # invalid gamma
    with pytest.raises(ValueError):
        __test_transform_base(pysptk.ignorm, 20, 0.1)


@pytest.mark.parametrize("src_order", [15, 20, 25, 30])
@pytest.mark.parametrize("dst_order", [15, 20, 25, 30])
@pytest.mark.parametrize("alpha", [0.35, 0.41, 0.5])
def test_freqt(src_order, dst_order, alpha):
    __test_conversion_base(pysptk.freqt, src_order, dst_order, alpha)


@pytest.mark.parametrize("src_order", [15, 20, 25, 30])
@pytest.mark.parametrize("src_alpha", [0.35, 0.41, 0.5])
@pytest.mark.parametrize("src_gamma", [-1.0, -0.5, 0.0])
@pytest.mark.parametrize("dst_order", [15, 20, 25, 30])
@pytest.mark.parametrize("dst_alpha", [0.35, 0.41, 0.5])
@pytest.mark.parametrize("dst_gamma", [-1.0, -0.5, 0.0])
def test_mgc2mgc(src_order, src_alpha, src_gamma, dst_order, dst_alpha, dst_gamma):
    np.random.seed(98765)
    src = np.random.rand(src_order + 1)
    dst = pysptk.mgc2mgc(src, src_alpha, src_gamma, dst_order, dst_alpha, dst_gamma)
    assert np.all(np.isfinite(dst))


def test_mgc2mgc_gamma():
    def __test(src_order, src_alpha, src_gamma, dst_order, dst_alpha, dst_gamma):
        np.random.seed(98765)
        src = np.random.rand(src_order + 1)
        dst = pysptk.mgc2mgc(src, src_alpha, src_gamma, dst_order, dst_alpha, dst_gamma)
        assert np.all(np.isfinite(dst))

    # invalid gamma
    with pytest.raises(ValueError):
        __test(20, 0.0, 0.1, 20, 0.0, 0.0)
    with pytest.raises(ValueError):
        __test(20, 0.0, 0.0, 20, 0.0, 0.1)


@pytest.mark.parametrize("order", [15, 20, 25, 30])
@pytest.mark.parametrize("alpha", [0.35, 0.41, 0.5])
@pytest.mark.parametrize("gamma", [-1.0, -0.5, 0.0])
@pytest.mark.parametrize("fftlen", [256, 512, 1024])
def test_mgc2sp(order, alpha, gamma, fftlen):
    np.random.seed(98765)
    src = np.random.rand(order + 1)
    dst = pysptk.mgc2sp(src, alpha, gamma, fftlen)
    assert len(dst) == (fftlen >> 1) + 1
    assert np.all(np.isfinite(dst))


@pytest.mark.parametrize("order", [15, 20, 25, 30])
@pytest.mark.parametrize("alpha", [0.35, 0.41, 0.5])
@pytest.mark.parametrize("gamma", [-1.0, -0.5, 0.0])
@pytest.mark.parametrize("fftlen", [256, 512, 1024])
def test_mgclsp2sp(order, alpha, gamma, fftlen):
    def __test(order, alpha, gamma, fftlen):
        np.random.seed(98765)
        src = np.random.rand(order + 1)
        dst = pysptk.mgclsp2sp(src, alpha, gamma, fftlen)
        assert len(dst) == (fftlen >> 1) + 1
        assert np.all(np.isfinite(dst))

    # TODO
    if gamma == 0.0:
        warn("Inf/-Inf wiil happens when gamma = 0.0")
        return

    __test(order, alpha, gamma, fftlen)


def test_mgclsp2sp_corner_case():
    def __test(order, alpha, gamma, fftlen):
        np.random.seed(98765)
        src = np.random.rand(order + 1)
        pysptk.mgclsp2sp(src, alpha, gamma, fftlen)

    # invalid gamma
    with pytest.raises(ValueError):
        __test(20, 0.0, 0.1, 256)

    # invalid fftlen
    with pytest.raises(ValueError):
        __test(20, 0.0, -0.1, 255)
    with pytest.raises(ValueError):
        __test(20, 0.0, -0.1, 257)


@pytest.mark.parametrize("order", [15, 20, 25, 30])
@pytest.mark.parametrize("alpha", [0.0, 0.35, 0.41])
@pytest.mark.parametrize("gamma", [-1.0, -0.5, 0.0])
def test_mgc2b(order, alpha, gamma):
    __test_transform_base(pysptk.mgc2b, order, alpha, gamma)


@pytest.mark.parametrize("order", [10, 20, 40, 50, 60])
@pytest.mark.parametrize("alpha", [0.35, 0.41])
@pytest.mark.parametrize("fftlen", [512, 1024, 2048])
def test_sp2mc(order, alpha, fftlen):
    np.random.seed(98765)
    sp = np.random.rand(int(fftlen // 2 + 1))
    mc = pysptk.sp2mc(sp, order, alpha)
    approx_sp = pysptk.mc2sp(mc, alpha, fftlen)
    # TODO: tolerance should be more carefully chosen
    assert np.allclose(sp, approx_sp, atol=0.9)


def test_mc2e():
    x = windowed_dummy_data(1024)
    mc = pysptk.mcep(x)
    assert pysptk.mc2e(mc) > 0
